import json

from sealy import (BfvEncryptionParametersBuilder, CoefficientModulus, Context,
                   DegreeType, KeyGenerator, PlainModulus, SecurityLevel)


def test_can_create_secret_key():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    secret_key = gen.secret_key()

    gen_2 = KeyGenerator(ctx)
    secret_key_2 = gen_2.secret_key()

    assert secret_key_2.as_bytes() != secret_key.as_bytes()


def test_can_create_public_key():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    gen.create_public_key()


def test_can_create_relin_key():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    gen.create_relinearization_key()


def test_can_create_galois_key():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.bfv(DegreeType(8192), SecurityLevel(128))
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 32))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    gen.create_galois_key()


def test_can_init_from_existing_secret_key():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    secret_key = gen.secret_key()

    gen_2 = KeyGenerator.from_secret_key(ctx, secret_key)
    secret_key_2 = gen_2.secret_key()

    assert secret_key_2.as_bytes() == secret_key.as_bytes()



from sealy import Plaintext


def test_can_create_and_destroy_plaintext():
    plaintext = Plaintext()
    del plaintext


def test_plaintext_coefficients_in_increasing_order():
    plaintext = Plaintext.from_hex_string("1234x^2 + 4321")

    assert plaintext.get_coefficient(0) == 0x4321
    assert plaintext.get_coefficient(1) == 0
    assert plaintext.get_coefficient(2) == 0x1234



# re-exporting the modules from sealy

from sealy.parameters import (BfvEncryptionParametersBuilder,
                              CkksEncryptionParametersBuilder)
from sealy.sealy import (AsymmetricComponents, BatchDecryptor, BatchEncryptor,
                         BFVDecimalEncoder, BFVEncoder, BFVEvaluator,
                         Ciphertext, CiphertextBatchArray, CKKSBatchEncoder,
                         CKKSBatchEvaluator, CKKSEncoder, CKKSEvaluator,
                         CoefficientModulus, Context, ContextData, Decryptor,
                         DegreeType, EncryptionParameters, Encryptor,
                         KeyGenerator, MemoryPool, Modulus, PlainModulus,
                         Plaintext, PlaintextBatchArray, PolynomialArray,
                         SchemeType, SecurityLevel)

__all__ = [
    "BfvEncryptionParametersBuilder",
    "CkksEncryptionParametersBuilder",
    "CoefficientModulus",
    "Context",
    "ContextData",
    "DegreeType",
    "EncryptionParameters",
    "KeyGenerator",
    "Modulus",
    "PlainModulus",
    "SchemeType",
    "SecurityLevel",
    "Plaintext",
    "MemoryPool",
    "Ciphertext",
    "PolynomialArray",
    "BFVEncoder",
    "BFVDecimalEncoder",
    "Encryptor",
    "AsymmetricComponents",
    "Decryptor",
    "BFVEvaluator",
    "CKKSEvaluator",
    "CKKSEncoder",
    "BatchEncryptor",
    "BatchDecryptor",
    "CiphertextBatchArray",
    "PlaintextBatchArray",
    "CKKSBatchEncoder",
    "CKKSBatchEvaluator",
]



from sealy import CoefficientModulus, PlainModulus, SecurityLevel
from sealy.parameters import (BfvEncryptionParametersBuilder,
                              CkksEncryptionParametersBuilder, DegreeType,
                              SchemeType)


def test_can_create_plain_modulus():
    modulus = PlainModulus.batching(DegreeType(1024), 20)
    assert modulus.get_value() == 1038337


def test_can_create_default_coefficient_modulus():
    modulus = CoefficientModulus.bfv(DegreeType(1024), SecurityLevel(128))
    assert len(modulus) == 1
    assert modulus[0].get_value() == 132120577

    modulus = CoefficientModulus.bfv(DegreeType(1024), SecurityLevel(192))
    assert len(modulus) == 1
    assert modulus[0].get_value() == 520193

    modulus = CoefficientModulus.bfv(DegreeType(1024), SecurityLevel(256))
    assert len(modulus) == 1
    assert modulus[0].get_value() == 12289


def test_can_create_custom_coefficient_modulus():
    modulus = CoefficientModulus.ckks(DegreeType(8192), [50, 30, 30, 50, 50])
    assert len(modulus) == 5
    assert modulus[0].get_value() == 1125899905744897
    assert modulus[1].get_value() == 1073643521
    assert modulus[2].get_value() == 1073692673
    assert modulus[3].get_value() == 1125899906629633
    assert modulus[4].get_value() == 1125899906826241


def test_can_roundtrip_security_level():
    for sec in [SecurityLevel(128), SecurityLevel(192), SecurityLevel(256)]:
        sec_2 = SecurityLevel(sec.get_value())
        assert sec == sec_2


def test_can_build_ckks_params():
    bit_sizes = [60, 40, 40, 60]
    modulus_chain = CoefficientModulus.ckks(DegreeType(1024), bit_sizes)

    params = (
        CkksEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(1024))
        .with_coefficient_modulus(modulus_chain)
        .build()
    )

    assert params.get_poly_modulus_degree() == 1024
    assert params.get_scheme() == SchemeType.ckks()
    assert len(params.get_coefficient_modulus()) == 4

    params = (
        CkksEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(1024))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
    ).build()

    modulus = params.get_coefficient_modulus()
    assert modulus[0].get_value() == 1125899905744897
    assert modulus[1].get_value() == 1073643521
    assert modulus[2].get_value() == 1073692673
    assert modulus[3].get_value() == 1125899906629633
    assert modulus[4].get_value() == 1125899906826241


def test_can_build_bfv_params():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(1024))
        .with_coefficient_modulus(
            CoefficientModulus.bfv(DegreeType(1024), SecurityLevel.default())
        )
        .with_plain_modulus_constant(1234)
    ).build()

    assert params.get_poly_modulus_degree() == 1024
    assert params.get_scheme() == SchemeType.bfv()
    assert params.get_plain_modulus().get_value() == 1234
    assert len(params.get_coefficient_modulus()) == 1
    assert params.get_coefficient_modulus()[0].get_value() == 132120577

    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(1024))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
    ).build()

    modulus = params.get_coefficient_modulus()
    assert modulus[0].get_value() == 1125899905744897
    assert modulus[1].get_value() == 1073643521
    assert modulus[2].get_value() == 1073692673
    assert modulus[3].get_value() == 1125899906629633
    assert modulus[4].get_value() == 1125899906826241



from typing import List

from sealy.sealy import DegreeType, EncryptionParameters, Modulus, SchemeType


class BfvEncryptionParametersBuilder:
    """Constructs BFV encryption"""

    def __init__(self) -> None:
        """Initializes a new instance of the BfvEncryptionParametersBuilder class."""
        self.poly_modulus_degree = None
        self.coeff_modulus = None
        self.plain_modulus = None

    def with_poly_modulus_degree(
        self, poly_modulus_degree: DegreeType
    ) -> "BfvEncryptionParametersBuilder":
        """Sets the polynomial degree of the underlying BFV scheme."""
        self.poly_modulus_degree = poly_modulus_degree
        return self

    def with_coefficient_modulus(
        self, coeff_modulus: List[Modulus]
    ) -> "BfvEncryptionParametersBuilder":
        """Sets the coefficient modulus for the encryption scheme."""
        self.coeff_modulus = coeff_modulus
        return self

    def with_plain_modulus(
        self, plain_modulus: Modulus
    ) -> "BfvEncryptionParametersBuilder":
        """Sets the plaintext modulus for the encryption scheme."""
        self.plain_modulus = plain_modulus
        return self

    def with_plain_modulus_constant(
        self, plain_modulus: int
    ) -> "BfvEncryptionParametersBuilder":
        """Sets the plaintext modulus for the encryption scheme."""
        self.plain_modulus = Modulus(plain_modulus)
        return self

    def build(self) -> EncryptionParameters:
        """Builds a new instance of the BfvEncryptionParameters class."""
        params = EncryptionParameters(SchemeType.bfv())

        if self.poly_modulus_degree is not None:
            params.set_poly_modulus_degree(self.poly_modulus_degree)
        else:
            raise ValueError("poly_modulus_degree cannot be None")

        if self.coeff_modulus is not None:
            params.set_coefficient_modulus(self.coeff_modulus)
        else:
            raise ValueError("coeff_modulus cannot be None")

        if self.plain_modulus is not None:
            if isinstance(self.plain_modulus, Modulus):
                params.set_plain_modulus(self.plain_modulus)
            elif isinstance(self.plain_modulus, int):
                params.set_plain_modulus_constant(self.plain_modulus)
            else:
                raise ValueError("plain_modulus must be Modulus or int")
        else:
            raise ValueError("plain_modulus cannot be None")

        return params


class CkksEncryptionParametersBuilder:
    """Constructs CKKS encryption"""

    def __init__(self) -> None:
        """Initializes a new instance of the CkksEncryptionParametersBuilder class."""
        self.poly_modulus_degree = None
        self.coeff_modulus = None

    def with_poly_modulus_degree(
        self, poly_modulus_degree: DegreeType
    ) -> "CkksEncryptionParametersBuilder":
        """Sets the polynomial degree of the underlying CKKS scheme."""
        self.poly_modulus_degree = poly_modulus_degree
        return self

    def with_coefficient_modulus(
        self, coeff_modulus: List[Modulus]
    ) -> "CkksEncryptionParametersBuilder":
        """Sets the coefficient modulus for the encryption scheme."""
        self.coeff_modulus = coeff_modulus
        return self

    def build(self) -> EncryptionParameters:
        """Builds a new instance of the CkksEncryptionParameters class."""
        params = EncryptionParameters(SchemeType.ckks())

        if self.poly_modulus_degree is not None:
            params.set_poly_modulus_degree(self.poly_modulus_degree)
        else:
            raise ValueError("poly_modulus_degree cannot be None")

        if self.coeff_modulus is not None:
            params.set_coefficient_modulus(self.coeff_modulus)
        else:
            raise ValueError("coeff_modulus cannot be None")

        return params



from sealy import (BfvEncryptionParametersBuilder, CoefficientModulus, Context,
                   DegreeType, SecurityLevel)


def test_can_create_and_drop_context():
    # Create encryption parameters
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(1024))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(1024), [60, 40, 40, 60])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    # Create context
    ctx = Context(params, False, SecurityLevel(128))

    # Drop context
    del ctx



import random
from typing import List

import pytest
from sealy import (BatchDecryptor, BatchEncryptor, CiphertextBatchArray,
                   CKKSBatchEncoder, CKKSBatchEvaluator,
                   CkksEncryptionParametersBuilder, CoefficientModulus,
                   Context, DegreeType, KeyGenerator, SecurityLevel)


def generate_random_tensor(size):
    return [random.uniform(0.0, 1.0) for _ in range(size)]


def average_ciphertexts(
    ctx: Context,
    encoder: CKKSBatchEncoder,
    ciphertexts: List[CiphertextBatchArray],
    size: int,
):
    evaluator = CKKSBatchEvaluator(ctx)

    cipher = evaluator.add_many(ciphertexts)

    fraction = 1.0 / len(ciphertexts)
    fraction = [fraction] * size
    fraction = encoder.encode(fraction)

    return evaluator.multiply_plain(cipher, fraction)


def average_plaintexts(plaintexts):
    avg = [0.0] * len(plaintexts[0])
    for tensor in plaintexts:
        for i, val in enumerate(tensor):
            avg[i] += val
    return [val / len(plaintexts) for val in avg]


@pytest.fixture
def context() -> Context:
    degree = DegreeType(8192)
    security_level = SecurityLevel(128)
    bit_sizes = [60, 40, 40, 60]

    expand_mod_chain = False
    modulus_chain = CoefficientModulus.create(degree, bit_sizes)
    encryption_parameters = (
        CkksEncryptionParametersBuilder()
        .with_poly_modulus_degree(degree)
        .with_coefficient_modulus(modulus_chain)
        .build()
    )

    return Context(encryption_parameters, expand_mod_chain, security_level)


@pytest.fixture
def key_generator(context: Context):
    return KeyGenerator(context)


@pytest.fixture
def encoder(context: Context):
    return CKKSBatchEncoder(context, 2**40)


@pytest.fixture
def encryptor(context: Context, key_generator: KeyGenerator):
    public_key = key_generator.create_public_key()
    return BatchEncryptor(context, public_key)


@pytest.fixture
def decryptor(context: Context, key_generator: KeyGenerator):
    private_key = key_generator.secret_key()
    return BatchDecryptor(context, private_key)


def test_average_ciphertexts(
    context: Context,
    encoder: CKKSBatchEncoder,
    encryptor: BatchEncryptor,
    decryptor: BatchDecryptor,
):
    client_1_gradients = generate_random_tensor(11000)
    client_2_gradients = generate_random_tensor(11000)
    client_3_gradients = generate_random_tensor(11000)

    client_1_encoded_gradients = encoder.encode(client_1_gradients)
    client_2_encoded_gradients = encoder.encode(client_2_gradients)
    client_3_encoded_gradients = encoder.encode(client_3_gradients)

    client_1_encrypted_gradients = encryptor.encrypt(
        client_1_encoded_gradients
    )
    client_2_encrypted_gradients = encryptor.encrypt(
        client_2_encoded_gradients
    )
    client_3_encrypted_gradients = encryptor.encrypt(
        client_3_encoded_gradients
    )

    avg_truth = average_plaintexts(
        [client_1_gradients, client_2_gradients, client_3_gradients]
    )

    avg = average_ciphertexts(
        context,
        encoder,
        [
            client_1_encrypted_gradients,
            client_2_encrypted_gradients,
            client_3_encrypted_gradients,
        ],
        10,
    )

    avg_dec = decryptor.decrypt(avg)
    avg_plain = encoder.decode(avg_dec)

    avg_plain = avg_plain[:10]

    for t, p in zip(avg_truth, avg_plain):
        assert abs(t - p) < 1e-6



from sealy import (CKKSEncoder, CkksEncryptionParametersBuilder,
                   CoefficientModulus, Context, DegreeType, SecurityLevel)


def test_can_create_and_drop_ckks_encoder():
    params = (
        CkksEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [60, 40, 40, 60])
        )
        .build()
    )
    scale = 2.0**40

    ctx = Context(params, False, SecurityLevel(128))
    encoder = CKKSEncoder(ctx, scale)
    del encoder


def test_can_get_slots_ckks_encoder():
    params = (
        CkksEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [60, 40, 40, 60])
        )
        .build()
    )
    scale = 2.0**40

    ctx = Context(params, False, SecurityLevel(128))
    encoder = CKKSEncoder(ctx, scale)

    assert encoder.get_slot_count() == 4096


def test_can_get_encode_and_decode_float():
    params = (
        CkksEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [60, 40, 40, 60])
        )
        .build()
    )
    scale = 2.0**40

    ctx = Context(params, False, SecurityLevel(128))
    encoder = CKKSEncoder(ctx, scale)

    data = [i / 10 for i in range(encoder.get_slot_count())]

    plaintext = encoder.encode(data)
    data_2 = encoder.decode(plaintext)

    # assert float array with 1e-6 precision
    assert all(abs(a - b) < 1e-6 for a, b in zip(data, data_2))



from sealy import (BfvEncryptionParametersBuilder, CoefficientModulus, Context,
                   DegreeType, Encryptor, KeyGenerator, SecurityLevel)
from sealy.sealy import BFVEncoder, Decryptor, PlainModulus


def test_can_encrypt_and_decrypt_unsigned():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))

    gen = KeyGenerator(ctx)

    encoder = BFVEncoder(ctx)

    data = [i for i in range(encoder.get_slot_count())]

    plaintext = encoder.encode(data)

    public_key = gen.create_public_key()
    secret_key = gen.secret_key()

    encryptor = Encryptor(ctx, public_key)
    decryptor = Decryptor(ctx, secret_key)

    # Asymmetric test
    ciphertext = encryptor.encrypt(plaintext)
    decrypted = decryptor.decrypt(ciphertext)
    data_2 = encoder.decode(decrypted)
    assert data == data_2


def test_can_encrypt_and_decrypt_from_return_components():

    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))

    gen = KeyGenerator(ctx)

    encoder = BFVEncoder(ctx)

    data = [i for i in range(encoder.get_slot_count())]

    plaintext = encoder.encode(data)

    public_key = gen.create_public_key()
    secret_key = gen.secret_key()

    encryptor = Encryptor(ctx, public_key)
    decryptor = Decryptor(ctx, secret_key)

    # Asymmetric test
    ciphertext = encryptor.encrypt_return_components(plaintext)[0]
    decrypted = decryptor.decrypt(ciphertext)
    data_2 = encoder.decode(decrypted)
    assert data == data_2

    # Symmetric test
    ciphertext = encryptor.encrypt_return_components(plaintext)[0]
    decrypted = decryptor.decrypt(ciphertext)
    data_2 = encoder.decode(decrypted)
    assert data == data_2



# ################################################################
# Copyright (c) 2020-2020, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under both the BSD-style license (found in the
# LICENSE file in the root directory of this source tree) and the GPLv2 (found
# in the COPYING file in the root directory of this source tree).
# You may select, at your option, one of the above-listed licenses.
# ##########################################################################

import argparse
import glob
import json
import os
import time
import pickle as pk
import subprocess
import urllib.request


GITHUB_API_PR_URL = "https://api.github.com/repos/facebook/zstd/pulls?state=open"
GITHUB_URL_TEMPLATE = "https://github.com/{}/zstd"
MASTER_BUILD = {"user": "facebook", "branch": "dev", "hash": None}

# check to see if there are any new PRs every minute
DEFAULT_MAX_API_CALL_FREQUENCY_SEC = 60
PREVIOUS_PRS_FILENAME = "prev_prs.pk"

# Not sure what the threshold for triggering alarms should be
# 1% regression sounds like a little too sensitive but the desktop
# that I'm running it on is pretty stable so I think this is fine
CSPEED_REGRESSION_TOLERANCE = 0.01
DSPEED_REGRESSION_TOLERANCE = 0.01


def get_new_open_pr_builds(prev_state=True):
    prev_prs = None
    if os.path.exists(PREVIOUS_PRS_FILENAME):
        with open(PREVIOUS_PRS_FILENAME, "rb") as f:
            prev_prs = pk.load(f)
    data = json.loads(urllib.request.urlopen(GITHUB_API_PR_URL).read().decode("utf-8"))
    prs = {
        d["url"]: {
            "user": d["user"]["login"],
            "branch": d["head"]["ref"],
            "hash": d["head"]["sha"].strip(),
        }
        for d in data
    }
    with open(PREVIOUS_PRS_FILENAME, "wb") as f:
        pk.dump(prs, f)
    if not prev_state or prev_prs == None:
        return list(prs.values())
    return [pr for url, pr in prs.items() if url not in prev_prs or prev_prs[url] != pr]


def get_latest_hashes():
    tmp = subprocess.run(["git", "log", "-1"], stdout=subprocess.PIPE).stdout.decode(
        "utf-8"
    )
    sha1 = tmp.split("\n")[0].split(" ")[1]
    tmp = subprocess.run(
        ["git", "show", "{}^1".format(sha1)], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    sha2 = tmp.split("\n")[0].split(" ")[1]
    tmp = subprocess.run(
        ["git", "show", "{}^2".format(sha1)], stdout=subprocess.PIPE
    ).stdout.decode("utf-8")
    sha3 = "" if len(tmp) == 0 else tmp.split("\n")[0].split(" ")[1]
    return [sha1.strip(), sha2.strip(), sha3.strip()]


def get_builds_for_latest_hash():
    hashes = get_latest_hashes()
    for b in get_new_open_pr_builds(False):
        if b["hash"] in hashes:
            return [b]
    return []


def clone_and_build(build):
    if build["user"] != None:
        github_url = GITHUB_URL_TEMPLATE.format(build["user"])
        os.system(
            """
            rm -rf zstd-{user}-{sha} &&
            git clone {github_url} zstd-{user}-{sha} &&
            cd zstd-{user}-{sha} &&
            {checkout_command}
            make &&
            cd ../
        """.format(
                user=build["user"],
                github_url=github_url,
                sha=build["hash"],
                checkout_command="git checkout {} &&".format(build["hash"])
                if build["hash"] != None
                else "",
            )
        )
        return "zstd-{user}-{sha}/zstd".format(user=build["user"], sha=build["hash"])
    else:
        os.system("cd ../ && make && cd tests")
        return "../zstd"


def parse_benchmark_output(output):
    idx = [i for i, d in enumerate(output) if d == "MB/s"]
    return [float(output[idx[0] - 1]), float(output[idx[1] - 1])]


def benchmark_single(executable, level, filename):
    return parse_benchmark_output((
        subprocess.run(
            [executable, "-qb{}".format(level), filename], stderr=subprocess.PIPE
        )
        .stderr.decode("utf-8")
        .split(" ")
    ))


def benchmark_n(executable, level, filename, n):
    speeds_arr = [benchmark_single(executable, level, filename) for _ in range(n)]
    cspeed, dspeed = max(b[0] for b in speeds_arr), max(b[1] for b in speeds_arr)
    print(
        "Bench (executable={} level={} filename={}, iterations={}):\n\t[cspeed: {} MB/s, dspeed: {} MB/s]".format(
            os.path.basename(executable),
            level,
            os.path.basename(filename),
            n,
            cspeed,
            dspeed,
        )
    )
    return (cspeed, dspeed)


def benchmark(build, filenames, levels, iterations):
    executable = clone_and_build(build)
    return [
        [benchmark_n(executable, l, f, iterations) for f in filenames] for l in levels
    ]


def benchmark_dictionary_single(executable, filenames_directory, dictionary_filename, level, iterations):
    cspeeds, dspeeds = [], []
    for _ in range(iterations):
        output = subprocess.run([executable, "-qb{}".format(level), "-D", dictionary_filename, "-r", filenames_directory], stderr=subprocess.PIPE).stderr.decode("utf-8").split(" ")
        cspeed, dspeed = parse_benchmark_output(output)
        cspeeds.append(cspeed)
        dspeeds.append(dspeed)
    max_cspeed, max_dspeed = max(cspeeds), max(dspeeds)
    print(
        "Bench (executable={} level={} filenames_directory={}, dictionary_filename={}, iterations={}):\n\t[cspeed: {} MB/s, dspeed: {} MB/s]".format(
            os.path.basename(executable),
            level,
            os.path.basename(filenames_directory),
            os.path.basename(dictionary_filename),
            iterations,
            max_cspeed,
            max_dspeed,
        )
    )
    return (max_cspeed, max_dspeed)


def benchmark_dictionary(build, filenames_directory, dictionary_filename, levels, iterations):
    executable = clone_and_build(build)
    return [benchmark_dictionary_single(executable, filenames_directory, dictionary_filename, l, iterations) for l in levels]


def parse_regressions_and_labels(old_cspeed, new_cspeed, old_dspeed, new_dspeed, baseline_build, test_build):
    cspeed_reg = (old_cspeed - new_cspeed) / old_cspeed
    dspeed_reg = (old_dspeed - new_dspeed) / old_dspeed
    baseline_label = "{}:{} ({})".format(
        baseline_build["user"], baseline_build["branch"], baseline_build["hash"]
    )
    test_label = "{}:{} ({})".format(
        test_build["user"], test_build["branch"], test_build["hash"]
    )
    return cspeed_reg, dspeed_reg, baseline_label, test_label


def get_regressions(baseline_build, test_build, iterations, filenames, levels):
    old = benchmark(baseline_build, filenames, levels, iterations)
    new = benchmark(test_build, filenames, levels, iterations)
    regressions = []
    for j, level in enumerate(levels):
        for k, filename in enumerate(filenames):
            old_cspeed, old_dspeed = old[j][k]
            new_cspeed, new_dspeed = new[j][k]
            cspeed_reg, dspeed_reg, baseline_label, test_label = parse_regressions_and_labels(
                old_cspeed, new_cspeed, old_dspeed, new_dspeed, baseline_build, test_build
            )
            if cspeed_reg > CSPEED_REGRESSION_TOLERANCE:
                regressions.append(
                    "[COMPRESSION REGRESSION] (level={} filename={})\n\t{} -> {}\n\t{} -> {} ({:0.2f}%)".format(
                        level,
                        filename,
                        baseline_label,
                        test_label,
                        old_cspeed,
                        new_cspeed,
                        cspeed_reg * 100.0,
                    )
                )
            if dspeed_reg > DSPEED_REGRESSION_TOLERANCE:
                regressions.append(
                    "[DECOMPRESSION REGRESSION] (level={} filename={})\n\t{} -> {}\n\t{} -> {} ({:0.2f}%)".format(
                        level,
                        filename,
                        baseline_label,
                        test_label,
                        old_dspeed,
                        new_dspeed,
                        dspeed_reg * 100.0,
                    )
                )
    return regressions

def get_regressions_dictionary(baseline_build, test_build, filenames_directory, dictionary_filename, levels, iterations):
    old = benchmark_dictionary(baseline_build, filenames_directory, dictionary_filename, levels, iterations)
    new = benchmark_dictionary(test_build, filenames_directory, dictionary_filename, levels, iterations)
    regressions = []
    for j, level in enumerate(levels):
        old_cspeed, old_dspeed = old[j]
        new_cspeed, new_dspeed = new[j]
        cspeed_reg, dspeed_reg, baesline_label, test_label = parse_regressions_and_labels(
            old_cspeed, new_cspeed, old_dspeed, new_dspeed, baseline_build, test_build
        )
        if cspeed_reg > CSPEED_REGRESSION_TOLERANCE:
            regressions.append(
                "[COMPRESSION REGRESSION] (level={} filenames_directory={} dictionary_filename={})\n\t{} -> {}\n\t{} -> {} ({:0.2f}%)".format(
                    level,
                    filenames_directory,
                    dictionary_filename,
                    baseline_label,
                    test_label,
                    old_cspeed,
                    new_cspeed,
                    cspeed_reg * 100.0,
                )
            )
        if dspeed_reg > DSPEED_REGRESSION_TOLERANCE:
            regressions.append(
                "[DECOMPRESSION REGRESSION] (level={} filenames_directory={} dictionary_filename={})\n\t{} -> {}\n\t{} -> {} ({:0.2f}%)".format(
                    level,
                    filenames_directory,
                    dictionary_filename,
                    baseline_label,
                    test_label,
                    old_dspeed,
                    new_dspeed,
                    dspeed_reg * 100.0,
                )
            )
        return regressions


def main(filenames, levels, iterations, builds=None, emails=None, continuous=False, frequency=DEFAULT_MAX_API_CALL_FREQUENCY_SEC, dictionary_filename=None):
    if builds == None:
        builds = get_new_open_pr_builds()
    while True:
        for test_build in builds:
            if dictionary_filename == None:
                regressions = get_regressions(
                    MASTER_BUILD, test_build, iterations, filenames, levels
                )
            else:
                regressions = get_regressions_dictionary(
                    MASTER_BUILD, test_build, filenames, dictionary_filename, levels, iterations
                )
            body = "\n".join(regressions)
            if len(regressions) > 0:
                if emails != None:
                    os.system(
                        """
                        echo "{}" | mutt -s "[zstd regression] caused by new pr" {}
                    """.format(
                            body, emails
                        )
                    )
                    print("Emails sent to {}".format(emails))
                print(body)
        if not continuous:
            break
        time.sleep(frequency)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--directory", help="directory with files to benchmark", default="golden-compression")
    parser.add_argument("--levels", help="levels to test eg ('1,2,3')", default="1")
    parser.add_argument("--iterations", help="number of benchmark iterations to run", default="1")
    parser.add_argument("--emails", help="email addresses of people who will be alerted upon regression. Only for continuous mode", default=None)
    parser.add_argument("--frequency", help="specifies the number of seconds to wait before each successive check for new PRs in continuous mode", default=DEFAULT_MAX_API_CALL_FREQUENCY_SEC)
    parser.add_argument("--mode", help="'fastmode', 'onetime', 'current', or 'continuous' (see README.md for details)", default="current")
    parser.add_argument("--dict", help="filename of dictionary to use (when set, this dictioanry will be used to compress the files provided inside --directory)", default=None)

    args = parser.parse_args()
    filenames = args.directory
    levels = [int(l) for l in args.levels.split(",")]
    mode = args.mode
    iterations = int(args.iterations)
    emails = args.emails
    frequency = int(args.frequency)
    dictionary_filename = args.dict

    if dictionary_filename == None:
        filenames = glob.glob("{}/**".format(filenames))

    if (len(filenames) == 0):
        print("0 files found")
        quit()

    if mode == "onetime":
        main(filenames, levels, iterations, frequency=frequenc, dictionary_filename=dictionary_filename)
    elif mode == "current":
        builds = [{"user": None, "branch": "None", "hash": None}]
        main(filenames, levels, iterations, builds, frequency=frequency, dictionary_filename=dictionary_filename)
    elif mode == "fastmode":
        builds = [{"user": "facebook", "branch": "master", "hash": None}]
        main(filenames, levels, iterations, builds, frequency=frequency, dictionary_filename=dictionary_filename)
    else:
        main(filenames, levels, iterations, None, emails, True, frequency=frequency, dictionary_filename=dictionary_filename)



from typing import Tuple

from sealy import (BFVEncoder, BfvEncryptionParametersBuilder, BFVEvaluator,
                   CoefficientModulus, Context, Decryptor, DegreeType,
                   Encryptor, KeyGenerator, PlainModulus, SecurityLevel)


def run_bfv_test() -> (
    Tuple[
        "Decryptor", "BFVEncoder", "Encryptor", "BFVEvaluator", "KeyGenerator"
    ]
):
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 32))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    encoder = BFVEncoder(ctx)

    public_key = gen.create_public_key()
    secret_key = gen.secret_key()

    encryptor = Encryptor(ctx, public_key)
    decryptor = Decryptor(ctx, secret_key)
    evaluator = BFVEvaluator(ctx)

    return (decryptor, encoder, encryptor, evaluator, gen)


def make_vec(encoder):
    return [
        encoder.get_slot_count() // 2 - i
        for i in range(encoder.get_slot_count())
    ]


def make_small_vec(encoder):
    return [16 - i % 32 for i in range(encoder.get_slot_count())]


def test_can_create_and_destroy_evaluator():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 32))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    evaluator = BFVEvaluator(ctx)
    del evaluator


def test_can_negate():
    decryptor, encoder, encryptor, evaluator, _ = run_bfv_test()

    vec = make_vec(encoder)
    encoded = encoder.encode(vec)
    encrypted = encryptor.encrypt(encoded)

    result = evaluator.negate(encrypted)

    decrypted = decryptor.decrypt(result)

    decoded = encoder.decode(decrypted)

    assert len(vec) == len(decoded)
    for i in range(len(vec)):
        assert vec[i] == -decoded[i]


def test_can_add():
    decryptor, encoder, encryptor, evaluator, _ = run_bfv_test()

    a = make_vec(encoder)
    b = make_vec(encoder)

    encoded_a = encoder.encode(a)
    encoded_b = encoder.encode(b)

    encrypted_a = encryptor.encrypt(encoded_a)
    encrypted_b = encryptor.encrypt(encoded_b)

    result = evaluator.add(encrypted_a, encrypted_b)

    decrypted = decryptor.decrypt(result)

    decoded = encoder.decode(decrypted)

    assert len(a) == len(decoded)
    for i in range(len(a)):
        assert a[i] + b[i] == decoded[i]


def test_can_sub():
    decryptor, encoder, encryptor, evaluator, _ = run_bfv_test()

    a = make_vec(encoder)
    b = make_vec(encoder)

    encoded_a = encoder.encode(a)
    encoded_b = encoder.encode(b)

    encrypted_a = encryptor.encrypt(encoded_a)
    encrypted_b = encryptor.encrypt(encoded_b)

    result = evaluator.sub(encrypted_a, encrypted_b)

    decrypted = decryptor.decrypt(result)

    decoded = encoder.decode(decrypted)

    assert len(a) == len(decoded)
    for i in range(len(a)):
        assert a[i] - b[i] == decoded[i]


def test_can_multiply():
    decryptor, encoder, encryptor, evaluator, _ = run_bfv_test()

    a = make_small_vec(encoder)
    b = make_small_vec(encoder)

    encoded_a = encoder.encode(a)
    encoded_b = encoder.encode(b)

    encrypted_a = encryptor.encrypt(encoded_a)
    encrypted_b = encryptor.encrypt(encoded_b)

    result = evaluator.multiply(encrypted_a, encrypted_b)

    decrypted = decryptor.decrypt(result)

    decoded = encoder.decode(decrypted)

    assert len(a) == len(decoded)
    for i in range(len(a)):
        assert a[i] * b[i] == decoded[i]



#!/usr/bin/env python
#
# Copyright 2010 Google Inc. All Rights Reserved.

"""Runs program specified in the command line with the substituted PATH.

   This script is needed for to support building under Pulse which is unable
   to override the existing PATH variable.
"""

import os
import subprocess
import sys

SUBST_PATH_ENV_VAR_NAME = "SUBST_PATH"

def main():
  if SUBST_PATH_ENV_VAR_NAME in os.environ:
    os.environ["PATH"] = os.environ[SUBST_PATH_ENV_VAR_NAME]

  exit_code = subprocess.Popen(sys.argv[1:]).wait()

  # exit_code is negative (-signal) if the process has been terminated by
  # a signal. Returning negative exit code is not portable and so we return
  # 100 instead.
  if exit_code < 0:
    exit_code = 100

  sys.exit(exit_code)

if __name__ == "__main__":
  main()



from sealy import (BfvEncryptionParametersBuilder, Ciphertext,
                   CoefficientModulus, Context, DegreeType, PolynomialArray,
                   SecurityLevel)
from sealy.sealy import BFVEncoder, Encryptor, KeyGenerator, PlainModulus


def test_can_create_and_destroy_static_polynomial_array():
    poly_array = PolynomialArray()
    del poly_array


def test_can_create_polynomial_from_ciphertext():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    ciphertext = Ciphertext()
    poly_array = PolynomialArray.from_ciphertext(ctx, ciphertext)

    assert poly_array.is_reserved()


def test_polynomial_array_initially_not_reserved():
    poly_array = PolynomialArray()
    assert not poly_array.is_reserved()


def generate_ciphertext_example():
    coeff_modulus = CoefficientModulus.create(
        DegreeType(8192), [50, 30, 30, 50, 50]
    )
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(coeff_modulus)
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    gen = KeyGenerator(ctx)

    encoder = BFVEncoder(ctx)

    data = [i for i in range(encoder.get_slot_count())]

    plaintext = encoder.encode(data)

    public_key = gen.create_public_key()
    secret_key = gen.secret_key()

    encryptor = Encryptor(ctx, public_key)

    ciphertext, components = encryptor.encrypt_return_components(plaintext)

    return (
        ctx,
        coeff_modulus,
        public_key,
        ciphertext,
        components.get_u(),
        components.get_e(),
        components.get_r(),
    )


def test_correct_poly_array_sizes_from_ciphertext():
    ctx, _, public_key, ciphertext, u, e, r = generate_ciphertext_example()
    poly_array_public_key = PolynomialArray.from_public_key(ctx, public_key)
    poly_array_ciphertext = PolynomialArray.from_ciphertext(ctx, ciphertext)

    poly_array_ciphertext_encoded = poly_array_ciphertext.as_ints()
    poly_array_public_key_encoded = poly_array_public_key.as_ints()
    u_encoded = u.as_ints()
    e_encoded = e.as_ints()

    assert poly_array_public_key.is_reserved()
    assert poly_array_ciphertext.is_reserved()
    assert u.is_reserved()
    assert e.is_reserved()

    # Ciphertext size checks
    assert (
        poly_array_ciphertext.get_num_polynomials()
        == ciphertext.get_num_polynomials()
    )
    assert poly_array_ciphertext.get_poly_modulus_degree() == 8192
    assert (
        poly_array_ciphertext.get_coeff_modulus_size()
        == ciphertext.get_coeff_modulus_size()
    )

    assert (
        len(poly_array_ciphertext_encoded)
        == poly_array_ciphertext.get_num_polynomials()
        * poly_array_ciphertext.get_poly_modulus_degree()
        * poly_array_ciphertext.get_coeff_modulus_size()
    )

    # Public key
    assert poly_array_public_key.get_num_polynomials() == 2
    assert poly_array_public_key.get_poly_modulus_degree() == 8192
    assert poly_array_public_key.get_coeff_modulus_size() == 4

    assert (
        len(poly_array_public_key_encoded)
        == poly_array_public_key.get_num_polynomials()
        * poly_array_public_key.get_poly_modulus_degree()
        * poly_array_public_key.get_coeff_modulus_size()
    )

    # u
    assert u.get_num_polynomials() == 1
    assert u.get_poly_modulus_degree() == 8192
    assert u.get_coeff_modulus_size() == 4
    assert (
        len(u_encoded)
        == poly_array_public_key.get_poly_modulus_degree()
        * poly_array_ciphertext.get_coeff_modulus_size()
    )

    # e
    assert e.get_num_polynomials() == 2
    assert e.get_poly_modulus_degree() == 8192
    assert e.get_coeff_modulus_size() == 4
    assert (
        len(e_encoded)
        == ciphertext.get_num_polynomials()
        * poly_array_public_key.get_poly_modulus_degree()
        * ciphertext.get_coeff_modulus_size()
    )

    # r
    assert r.size() == 8192


def test_multiprecision_and_back_is_identity():
    ctx, _, _, ciphertext, _, _, _ = generate_ciphertext_example()
    poly_array = PolynomialArray.from_ciphertext(ctx, ciphertext)

    poly_array_encoded_original = poly_array.as_ints()

    poly_array.to_multiprecision()
    poly_array.to_rns()

    poly_array_encoded_round_trip = poly_array.as_ints()

    assert poly_array_encoded_original == poly_array_encoded_round_trip



from sealy import (BfvEncryptionParametersBuilder, CoefficientModulus, Context,
                   DegreeType, Encryptor, KeyGenerator, SecurityLevel)


def test_can_create_and_drop_encryptor():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus_constant(1234)
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))

    gen = KeyGenerator(ctx)

    public_key = gen.create_public_key()

    encryptor = Encryptor(ctx, public_key)

    del encryptor



#!/usr/bin/env python3

# ################################################################
# Copyright (c) 2016-2020, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under both the BSD-style license (found in the
# LICENSE file in the root directory of this source tree) and the GPLv2 (found
# in the COPYING file in the root directory of this source tree).
# You may select, at your option, one of the above-listed licenses.
# ################################################################

import datetime
import enum
import glob
import os
import sys

YEAR = datetime.datetime.now().year

YEAR_STR = str(YEAR)

ROOT = os.path.join(os.path.dirname(__file__), "..")

RELDIRS = [
    "doc",
    "examples",
    "lib",
    "programs",
    "tests",
]

DIRS = [os.path.join(ROOT, d) for d in RELDIRS]

class File(enum.Enum):
    C = 1
    H = 2
    MAKE = 3
    PY = 4

SUFFIX = {
    File.C: ".c",
    File.H: ".h",
    File.MAKE: "Makefile",
    File.PY: ".py",
}

# License should certainly be in the first 10 KB.
MAX_BYTES = 10000
MAX_LINES = 50

LICENSE_LINES = [
    "This source code is licensed under both the BSD-style license (found in the",
    "LICENSE file in the root directory of this source tree) and the GPLv2 (found",
    "in the COPYING file in the root directory of this source tree).",
    "You may select, at your option, one of the above-listed licenses.",
]

COPYRIGHT_EXCEPTIONS = {
    # From zstdmt
    "threading.c",
    "threading.h",
    # From divsufsort
    "divsufsort.c",
    "divsufsort.h",
}

LICENSE_EXCEPTIONS = {
    # From divsufsort
    "divsufsort.c",
    "divsufsort.h",
}


def valid_copyright(lines):
    for line in lines:
        line = line.strip()
        if "Copyright" not in line:
            continue
        if "present" in line:
            return (False, f"Copyright line '{line}' contains 'present'!")
        if "Facebook, Inc" not in line:
            return (False, f"Copyright line '{line}' does not contain 'Facebook, Inc'")
        if YEAR_STR not in line:
            return (False, f"Copyright line '{line}' does not contain {YEAR}")
        if " (c) " not in line:
            return (False, f"Copyright line '{line}' does not contain ' (c) '!")
        return (True, "")
    return (False, "Copyright not found!")


def valid_license(lines):
    for b in range(len(lines)):
        if LICENSE_LINES[0] not in lines[b]:
            continue
        for l in range(len(LICENSE_LINES)):
            if LICENSE_LINES[l] not in lines[b + l]:
                message = f"""Invalid license line found starting on line {b + l}!
Expected: '{LICENSE_LINES[l]}'
Actual: '{lines[b + l]}'"""
                return (False, message)
        return (True, "")
    return (False, "License not found!")


def valid_file(filename):
    with open(filename, "r") as f:
        lines = f.readlines(MAX_BYTES)
    lines = lines[:min(len(lines), MAX_LINES)]
                
    ok = True
    if os.path.basename(filename) not in COPYRIGHT_EXCEPTIONS:
        c_ok, c_msg = valid_copyright(lines)
        if not c_ok:
            print(f"{filename}: {c_msg}")
            ok = False
    if os.path.basename(filename) not in LICENSE_EXCEPTIONS:
        l_ok, l_msg = valid_license(lines)
        if not l_ok:
            print(f"{filename}: {l_msg}")
            ok = False
    return ok


def main():
    invalid_files = []
    for directory in DIRS:
        for suffix in SUFFIX.values():
            files = set(glob.glob(f"{directory}/*{suffix}"))
            files |= set(glob.glob(f"{directory}/**/*{suffix}"))
            for filename in files:
                if not valid_file(filename):
                    invalid_files.append(filename)
    if len(invalid_files) > 0:
        print(f"Invalid files: {invalid_files}")
    else:
        print("Pass!")
    return len(invalid_files)

if __name__ == "__main__":
    sys.exit(main())


#!/usr/bin/env python
#
# Copyright 2007 Neal Norwitz
# Portions Copyright 2007 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generate an Abstract Syntax Tree (AST) for C++."""

# FIXME:
#  * Tokens should never be exported, need to convert to Nodes
#    (return types, parameters, etc.)
#  * Handle static class data for templatized classes
#  * Handle casts (both C++ and C-style)
#  * Handle conditions and loops (if/else, switch, for, while/do)
#
# TODO much, much later:
#  * Handle #define
#  * exceptions


try:
  # Python 3.x
  import builtins
except ImportError:
  # Python 2.x
  import __builtin__ as builtins

import collections
import sys
import traceback

from cpp import keywords
from cpp import tokenize
from cpp import utils


if not hasattr(builtins, 'reversed'):
  # Support Python 2.3 and earlier.
  def reversed(seq):
    for i in range(len(seq)-1, -1, -1):
      yield seq[i]

if not hasattr(builtins, 'next'):
  # Support Python 2.5 and earlier.
  def next(obj):
    return obj.next()


VISIBILITY_PUBLIC, VISIBILITY_PROTECTED, VISIBILITY_PRIVATE = range(3)

FUNCTION_NONE = 0x00
FUNCTION_CONST = 0x01
FUNCTION_VIRTUAL = 0x02
FUNCTION_PURE_VIRTUAL = 0x04
FUNCTION_CTOR = 0x08
FUNCTION_DTOR = 0x10
FUNCTION_ATTRIBUTE = 0x20
FUNCTION_UNKNOWN_ANNOTATION = 0x40
FUNCTION_THROW = 0x80
FUNCTION_OVERRIDE = 0x100

"""
These are currently unused.  Should really handle these properly at some point.

TYPE_MODIFIER_INLINE   = 0x010000
TYPE_MODIFIER_EXTERN   = 0x020000
TYPE_MODIFIER_STATIC   = 0x040000
TYPE_MODIFIER_CONST    = 0x080000
TYPE_MODIFIER_REGISTER = 0x100000
TYPE_MODIFIER_VOLATILE = 0x200000
TYPE_MODIFIER_MUTABLE  = 0x400000

TYPE_MODIFIER_MAP = {
    'inline': TYPE_MODIFIER_INLINE,
    'extern': TYPE_MODIFIER_EXTERN,
    'static': TYPE_MODIFIER_STATIC,
    'const': TYPE_MODIFIER_CONST,
    'register': TYPE_MODIFIER_REGISTER,
    'volatile': TYPE_MODIFIER_VOLATILE,
    'mutable': TYPE_MODIFIER_MUTABLE,
    }
"""

_INTERNAL_TOKEN = 'internal'
_NAMESPACE_POP = 'ns-pop'


# TODO(nnorwitz): use this as a singleton for templated_types, etc
# where we don't want to create a new empty dict each time.  It is also const.
class _NullDict(object):
  __contains__ = lambda self: False
  keys = values = items = iterkeys = itervalues = iteritems = lambda self: ()


# TODO(nnorwitz): move AST nodes into a separate module.
class Node(object):
  """Base AST node."""

  def __init__(self, start, end):
    self.start = start
    self.end = end

  def IsDeclaration(self):
    """Returns bool if this node is a declaration."""
    return False

  def IsDefinition(self):
    """Returns bool if this node is a definition."""
    return False

  def IsExportable(self):
    """Returns bool if this node exportable from a header file."""
    return False

  def Requires(self, node):
    """Does this AST node require the definition of the node passed in?"""
    return False

  def XXX__str__(self):
    return self._StringHelper(self.__class__.__name__, '')

  def _StringHelper(self, name, suffix):
    if not utils.DEBUG:
      return '%s(%s)' % (name, suffix)
    return '%s(%d, %d, %s)' % (name, self.start, self.end, suffix)

  def __repr__(self):
    return str(self)


class Define(Node):
  def __init__(self, start, end, name, definition):
    Node.__init__(self, start, end)
    self.name = name
    self.definition = definition

  def __str__(self):
    value = '%s %s' % (self.name, self.definition)
    return self._StringHelper(self.__class__.__name__, value)


class Include(Node):
  def __init__(self, start, end, filename, system):
    Node.__init__(self, start, end)
    self.filename = filename
    self.system = system

  def __str__(self):
    fmt = '"%s"'
    if self.system:
      fmt = '<%s>'
    return self._StringHelper(self.__class__.__name__, fmt % self.filename)


class Goto(Node):
  def __init__(self, start, end, label):
    Node.__init__(self, start, end)
    self.label = label

  def __str__(self):
    return self._StringHelper(self.__class__.__name__, str(self.label))


class Expr(Node):
  def __init__(self, start, end, expr):
    Node.__init__(self, start, end)
    self.expr = expr

  def Requires(self, node):
    # TODO(nnorwitz): impl.
    return False

  def __str__(self):
    return self._StringHelper(self.__class__.__name__, str(self.expr))


class Return(Expr):
  pass


class Delete(Expr):
  pass


class Friend(Expr):
  def __init__(self, start, end, expr, namespace):
    Expr.__init__(self, start, end, expr)
    self.namespace = namespace[:]


class Using(Node):
  def __init__(self, start, end, names):
    Node.__init__(self, start, end)
    self.names = names

  def __str__(self):
    return self._StringHelper(self.__class__.__name__, str(self.names))


class Parameter(Node):
  def __init__(self, start, end, name, parameter_type, default):
    Node.__init__(self, start, end)
    self.name = name
    self.type = parameter_type
    self.default = default

  def Requires(self, node):
    # TODO(nnorwitz): handle namespaces, etc.
    return self.type.name == node.name

  def __str__(self):
    name = str(self.type)
    suffix = '%s %s' % (name, self.name)
    if self.default:
      suffix += ' = ' + ''.join([d.name for d in self.default])
    return self._StringHelper(self.__class__.__name__, suffix)


class _GenericDeclaration(Node):
  def __init__(self, start, end, name, namespace):
    Node.__init__(self, start, end)
    self.name = name
    self.namespace = namespace[:]

  def FullName(self):
    prefix = ''
    if self.namespace and self.namespace[-1]:
      prefix = '::'.join(self.namespace) + '::'
    return prefix + self.name

  def _TypeStringHelper(self, suffix):
    if self.namespace:
      names = [n or '<anonymous>' for n in self.namespace]
      suffix += ' in ' + '::'.join(names)
    return self._StringHelper(self.__class__.__name__, suffix)


# TODO(nnorwitz): merge with Parameter in some way?
class VariableDeclaration(_GenericDeclaration):
  def __init__(self, start, end, name, var_type, initial_value, namespace):
    _GenericDeclaration.__init__(self, start, end, name, namespace)
    self.type = var_type
    self.initial_value = initial_value

  def Requires(self, node):
    # TODO(nnorwitz): handle namespaces, etc.
    return self.type.name == node.name

  def ToString(self):
    """Return a string that tries to reconstitute the variable decl."""
    suffix = '%s %s' % (self.type, self.name)
    if self.initial_value:
      suffix += ' = ' + self.initial_value
    return suffix

  def __str__(self):
    return self._StringHelper(self.__class__.__name__, self.ToString())


class Typedef(_GenericDeclaration):
  def __init__(self, start, end, name, alias, namespace):
    _GenericDeclaration.__init__(self, start, end, name, namespace)
    self.alias = alias

  def IsDefinition(self):
    return True

  def IsExportable(self):
    return True

  def Requires(self, node):
    # TODO(nnorwitz): handle namespaces, etc.
    name = node.name
    for token in self.alias:
      if token is not None and name == token.name:
        return True
    return False

  def __str__(self):
    suffix = '%s, %s' % (self.name, self.alias)
    return self._TypeStringHelper(suffix)


class _NestedType(_GenericDeclaration):
  def __init__(self, start, end, name, fields, namespace):
    _GenericDeclaration.__init__(self, start, end, name, namespace)
    self.fields = fields

  def IsDefinition(self):
    return True

  def IsExportable(self):
    return True

  def __str__(self):
    suffix = '%s, {%s}' % (self.name, self.fields)
    return self._TypeStringHelper(suffix)


class Union(_NestedType):
  pass


class Enum(_NestedType):
  pass


class Class(_GenericDeclaration):
  def __init__(self, start, end, name, bases, templated_types, body, namespace):
    _GenericDeclaration.__init__(self, start, end, name, namespace)
    self.bases = bases
    self.body = body
    self.templated_types = templated_types

  def IsDeclaration(self):
    return self.bases is None and self.body is None

  def IsDefinition(self):
    return not self.IsDeclaration()

  def IsExportable(self):
    return not self.IsDeclaration()

  def Requires(self, node):
    # TODO(nnorwitz): handle namespaces, etc.
    if self.bases:
      for token_list in self.bases:
        # TODO(nnorwitz): bases are tokens, do name comparison.
        for token in token_list:
          if token.name == node.name:
            return True
    # TODO(nnorwitz): search in body too.
    return False

  def __str__(self):
    name = self.name
    if self.templated_types:
      name += '<%s>' % self.templated_types
    suffix = '%s, %s, %s' % (name, self.bases, self.body)
    return self._TypeStringHelper(suffix)


class Struct(Class):
  pass


class Function(_GenericDeclaration):
  def __init__(self, start, end, name, return_type, parameters,
               modifiers, templated_types, body, namespace):
    _GenericDeclaration.__init__(self, start, end, name, namespace)
    converter = TypeConverter(namespace)
    self.return_type = converter.CreateReturnType(return_type)
    self.parameters = converter.ToParameters(parameters)
    self.modifiers = modifiers
    self.body = body
    self.templated_types = templated_types

  def IsDeclaration(self):
    return self.body is None

  def IsDefinition(self):
    return self.body is not None

  def IsExportable(self):
    if self.return_type and 'static' in self.return_type.modifiers:
      return False
    return None not in self.namespace

  def Requires(self, node):
    if self.parameters:
      # TODO(nnorwitz): parameters are tokens, do name comparison.
      for p in self.parameters:
        if p.name == node.name:
          return True
    # TODO(nnorwitz): search in body too.
    return False

  def __str__(self):
    # TODO(nnorwitz): add templated_types.
    suffix = ('%s %s(%s), 0x%02x, %s' %
              (self.return_type, self.name, self.parameters,
               self.modifiers, self.body))
    return self._TypeStringHelper(suffix)


class Method(Function):
  def __init__(self, start, end, name, in_class, return_type, parameters,
               modifiers, templated_types, body, namespace):
    Function.__init__(self, start, end, name, return_type, parameters,
                      modifiers, templated_types, body, namespace)
    # TODO(nnorwitz): in_class could also be a namespace which can
    # mess up finding functions properly.
    self.in_class = in_class


class Type(_GenericDeclaration):
  """Type used for any variable (eg class, primitive, struct, etc)."""

  def __init__(self, start, end, name, templated_types, modifiers,
               reference, pointer, array):
    """
        Args:
          name: str name of main type
          templated_types: [Class (Type?)] template type info between <>
          modifiers: [str] type modifiers (keywords) eg, const, mutable, etc.
          reference, pointer, array: bools
        """
    _GenericDeclaration.__init__(self, start, end, name, [])
    self.templated_types = templated_types
    if not name and modifiers:
      self.name = modifiers.pop()
    self.modifiers = modifiers
    self.reference = reference
    self.pointer = pointer
    self.array = array

  def __str__(self):
    prefix = ''
    if self.modifiers:
      prefix = ' '.join(self.modifiers) + ' '
    name = str(self.name)
    if self.templated_types:
      name += '<%s>' % self.templated_types
    suffix = prefix + name
    if self.reference:
      suffix += '&'
    if self.pointer:
      suffix += '*'
    if self.array:
      suffix += '[]'
    return self._TypeStringHelper(suffix)

  # By definition, Is* are always False.  A Type can only exist in
  # some sort of variable declaration, parameter, or return value.
  def IsDeclaration(self):
    return False

  def IsDefinition(self):
    return False

  def IsExportable(self):
    return False


class TypeConverter(object):

  def __init__(self, namespace_stack):
    self.namespace_stack = namespace_stack

  def _GetTemplateEnd(self, tokens, start):
    count = 1
    end = start
    while 1:
      token = tokens[end]
      end += 1
      if token.name == '<':
        count += 1
      elif token.name == '>':
        count -= 1
        if count == 0:
          break
    return tokens[start:end-1], end

  def ToType(self, tokens):
    """Convert [Token,...] to [Class(...), ] useful for base classes.
        For example, code like class Foo : public Bar<x, y> { ... };
        the "Bar<x, y>" portion gets converted to an AST.

        Returns:
          [Class(...), ...]
        """
    result = []
    name_tokens = []
    reference = pointer = array = False

    def AddType(templated_types):
      # Partition tokens into name and modifier tokens.
      names = []
      modifiers = []
      for t in name_tokens:
        if keywords.IsKeyword(t.name):
          modifiers.append(t.name)
        else:
          names.append(t.name)
      name = ''.join(names)
      if name_tokens:
        result.append(Type(name_tokens[0].start, name_tokens[-1].end,
                           name, templated_types, modifiers,
                           reference, pointer, array))
      del name_tokens[:]

    i = 0
    end = len(tokens)
    while i < end:
      token = tokens[i]
      if token.name == '<':
        new_tokens, new_end = self._GetTemplateEnd(tokens, i+1)
        AddType(self.ToType(new_tokens))
        # If there is a comma after the template, we need to consume
        # that here otherwise it becomes part of the name.
        i = new_end
        reference = pointer = array = False
      elif token.name == ',':
        AddType([])
        reference = pointer = array = False
      elif token.name == '*':
        pointer = True
      elif token.name == '&':
        reference = True
      elif token.name == '[':
        pointer = True
      elif token.name == ']':
        pass
      else:
        name_tokens.append(token)
      i += 1

    if name_tokens:
      # No '<' in the tokens, just a simple name and no template.
      AddType([])
    return result

  def DeclarationToParts(self, parts, needs_name_removed):
    name = None
    default = []
    if needs_name_removed:
      # Handle default (initial) values properly.
      for i, t in enumerate(parts):
        if t.name == '=':
          default = parts[i+1:]
          name = parts[i-1].name
          if name == ']' and parts[i-2].name == '[':
            name = parts[i-3].name
            i -= 1
          parts = parts[:i-1]
          break
      else:
        if parts[-1].token_type == tokenize.NAME:
          name = parts.pop().name
        else:
          # TODO(nnorwitz): this is a hack that happens for code like
          # Register(Foo<T>); where it thinks this is a function call
          # but it's actually a declaration.
          name = '???'
    modifiers = []
    type_name = []
    other_tokens = []
    templated_types = []
    i = 0
    end = len(parts)
    while i < end:
      p = parts[i]
      if keywords.IsKeyword(p.name):
        modifiers.append(p.name)
      elif p.name == '<':
        templated_tokens, new_end = self._GetTemplateEnd(parts, i+1)
        templated_types = self.ToType(templated_tokens)
        i = new_end - 1
        # Don't add a spurious :: to data members being initialized.
        next_index = i + 1
        if next_index < end and parts[next_index].name == '::':
          i += 1
      elif p.name in ('[', ']', '='):
        # These are handled elsewhere.
        other_tokens.append(p)
      elif p.name not in ('*', '&', '>'):
        # Ensure that names have a space between them.
        if (type_name and type_name[-1].token_type == tokenize.NAME and
                p.token_type == tokenize.NAME):
          type_name.append(tokenize.Token(tokenize.SYNTAX, ' ', 0, 0))
        type_name.append(p)
      else:
        other_tokens.append(p)
      i += 1
    type_name = ''.join([t.name for t in type_name])
    return name, type_name, templated_types, modifiers, default, other_tokens

  def ToParameters(self, tokens):
    if not tokens:
      return []

    result = []
    name = type_name = ''
    type_modifiers = []
    pointer = reference = array = False
    first_token = None
    default = []

    def AddParameter(end):
      if default:
        del default[0]  # Remove flag.
      parts = self.DeclarationToParts(type_modifiers, True)
      (name, type_name, templated_types, modifiers,
       unused_default, unused_other_tokens) = parts
      parameter_type = Type(first_token.start, first_token.end,
                            type_name, templated_types, modifiers,
                            reference, pointer, array)
      p = Parameter(first_token.start, end, name,
                    parameter_type, default)
      result.append(p)

    template_count = 0
    brace_count = 0
    for s in tokens:
      if not first_token:
        first_token = s

      # Check for braces before templates, as we can have unmatched '<>'
      # inside default arguments.
      if s.name == '{':
        brace_count += 1
      elif s.name == '}':
        brace_count -= 1
      if brace_count > 0:
        type_modifiers.append(s)
        continue

      if s.name == '<':
        template_count += 1
      elif s.name == '>':
        template_count -= 1
      if template_count > 0:
        type_modifiers.append(s)
        continue

      if s.name == ',':
        AddParameter(s.start)
        name = type_name = ''
        type_modifiers = []
        pointer = reference = array = False
        first_token = None
        default = []
      elif s.name == '*':
        pointer = True
      elif s.name == '&':
        reference = True
      elif s.name == '[':
        array = True
      elif s.name == ']':
        pass  # Just don't add to type_modifiers.
      elif s.name == '=':
        # Got a default value.  Add any value (None) as a flag.
        default.append(None)
      elif default:
        default.append(s)
      else:
        type_modifiers.append(s)
    AddParameter(tokens[-1].end)
    return result

  def CreateReturnType(self, return_type_seq):
    if not return_type_seq:
      return None
    start = return_type_seq[0].start
    end = return_type_seq[-1].end
    _, name, templated_types, modifiers, default, other_tokens = \
        self.DeclarationToParts(return_type_seq, False)
    names = [n.name for n in other_tokens]
    reference = '&' in names
    pointer = '*' in names
    array = '[' in names
    return Type(start, end, name, templated_types, modifiers,
                reference, pointer, array)

  def GetTemplateIndices(self, names):
    # names is a list of strings.
    start = names.index('<')
    end = len(names) - 1
    while end > 0:
      if names[end] == '>':
        break
      end -= 1
    return start, end+1

class AstBuilder(object):
  def __init__(self, token_stream, filename, in_class='', visibility=None,
               namespace_stack=[]):
    self.tokens = token_stream
    self.filename = filename
    # TODO(nnorwitz): use a better data structure (deque) for the queue.
    # Switching directions of the "queue" improved perf by about 25%.
    # Using a deque should be even better since we access from both sides.
    self.token_queue = []
    self.namespace_stack = namespace_stack[:]
    self.in_class = in_class
    if in_class is None:
      self.in_class_name_only = None
    else:
      self.in_class_name_only = in_class.split('::')[-1]
    self.visibility = visibility
    self.in_function = False
    self.current_token = None
    # Keep the state whether we are currently handling a typedef or not.
    self._handling_typedef = False

    self.converter = TypeConverter(self.namespace_stack)

  def HandleError(self, msg, token):
    printable_queue = list(reversed(self.token_queue[-20:]))
    sys.stderr.write('Got %s in %s @ %s %s\n' %
                     (msg, self.filename, token, printable_queue))

  def Generate(self):
    while 1:
      token = self._GetNextToken()
      if not token:
        break

      # Get the next token.
      self.current_token = token

      # Dispatch on the next token type.
      if token.token_type == _INTERNAL_TOKEN:
        if token.name == _NAMESPACE_POP:
          self.namespace_stack.pop()
        continue

      try:
        result = self._GenerateOne(token)
        if result is not None:
          yield result
      except:
        self.HandleError('exception', token)
        raise

  def _CreateVariable(self, pos_token, name, type_name, type_modifiers,
                      ref_pointer_name_seq, templated_types, value=None):
    reference = '&' in ref_pointer_name_seq
    pointer = '*' in ref_pointer_name_seq
    array = '[' in ref_pointer_name_seq
    var_type = Type(pos_token.start, pos_token.end, type_name,
                    templated_types, type_modifiers,
                    reference, pointer, array)
    return VariableDeclaration(pos_token.start, pos_token.end,
                               name, var_type, value, self.namespace_stack)

  def _GenerateOne(self, token):
    if token.token_type == tokenize.NAME:
      if (keywords.IsKeyword(token.name) and
          not keywords.IsBuiltinType(token.name)):
        if token.name == 'enum':
          # Pop the next token and only put it back if it's not
          # 'class'.  This allows us to support the two-token
          # 'enum class' keyword as if it were simply 'enum'.
          next = self._GetNextToken()
          if next.name != 'class':
            self._AddBackToken(next)

        method = getattr(self, 'handle_' + token.name)
        return method()
      elif token.name == self.in_class_name_only:
        # The token name is the same as the class, must be a ctor if
        # there is a paren.  Otherwise, it's the return type.
        # Peek ahead to get the next token to figure out which.
        next = self._GetNextToken()
        self._AddBackToken(next)
        if next.token_type == tokenize.SYNTAX and next.name == '(':
          return self._GetMethod([token], FUNCTION_CTOR, None, True)
        # Fall through--handle like any other method.

      # Handle data or function declaration/definition.
      syntax = tokenize.SYNTAX
      temp_tokens, last_token = \
          self._GetVarTokensUpToIgnoringTemplates(syntax,
                                                  '(', ';', '{', '[')
      temp_tokens.insert(0, token)
      if last_token.name == '(':
        # If there is an assignment before the paren,
        # this is an expression, not a method.
        expr = bool([e for e in temp_tokens if e.name == '='])
        if expr:
          new_temp = self._GetTokensUpTo(tokenize.SYNTAX, ';')
          temp_tokens.append(last_token)
          temp_tokens.extend(new_temp)
          last_token = tokenize.Token(tokenize.SYNTAX, ';', 0, 0)

      if last_token.name == '[':
        # Handle array, this isn't a method, unless it's an operator.
        # TODO(nnorwitz): keep the size somewhere.
        # unused_size = self._GetTokensUpTo(tokenize.SYNTAX, ']')
        temp_tokens.append(last_token)
        if temp_tokens[-2].name == 'operator':
          temp_tokens.append(self._GetNextToken())
        else:
          temp_tokens2, last_token = \
              self._GetVarTokensUpTo(tokenize.SYNTAX, ';')
          temp_tokens.extend(temp_tokens2)

      if last_token.name == ';':
        # Handle data, this isn't a method.
        parts = self.converter.DeclarationToParts(temp_tokens, True)
        (name, type_name, templated_types, modifiers, default,
         unused_other_tokens) = parts

        t0 = temp_tokens[0]
        names = [t.name for t in temp_tokens]
        if templated_types:
          start, end = self.converter.GetTemplateIndices(names)
          names = names[:start] + names[end:]
        default = ''.join([t.name for t in default])
        return self._CreateVariable(t0, name, type_name, modifiers,
                                    names, templated_types, default)
      if last_token.name == '{':
        self._AddBackTokens(temp_tokens[1:])
        self._AddBackToken(last_token)
        method_name = temp_tokens[0].name
        method = getattr(self, 'handle_' + method_name, None)
        if not method:
          # Must be declaring a variable.
          # TODO(nnorwitz): handle the declaration.
          return None
        return method()
      return self._GetMethod(temp_tokens, 0, None, False)
    elif token.token_type == tokenize.SYNTAX:
      if token.name == '~' and self.in_class:
        # Must be a dtor (probably not in method body).
        token = self._GetNextToken()
        # self.in_class can contain A::Name, but the dtor will only
        # be Name.  Make sure to compare against the right value.
        if (token.token_type == tokenize.NAME and
                token.name == self.in_class_name_only):
          return self._GetMethod([token], FUNCTION_DTOR, None, True)
      # TODO(nnorwitz): handle a lot more syntax.
    elif token.token_type == tokenize.PREPROCESSOR:
      # TODO(nnorwitz): handle more preprocessor directives.
      # token starts with a #, so remove it and strip whitespace.
      name = token.name[1:].lstrip()
      if name.startswith('include'):
        # Remove "include".
        name = name[7:].strip()
        assert name
        # Handle #include \<newline> "header-on-second-line.h".
        if name.startswith('\\'):
          name = name[1:].strip()
        assert name[0] in '<"', token
        assert name[-1] in '>"', token
        system = name[0] == '<'
        filename = name[1:-1]
        return Include(token.start, token.end, filename, system)
      if name.startswith('define'):
        # Remove "define".
        name = name[6:].strip()
        assert name
        value = ''
        for i, c in enumerate(name):
          if c.isspace():
            value = name[i:].lstrip()
            name = name[:i]
            break
        return Define(token.start, token.end, name, value)
      if name.startswith('if') and name[2:3].isspace():
        condition = name[3:].strip()
        if condition.startswith('0') or condition.startswith('(0)'):
          self._SkipIf0Blocks()
    return None

  def _GetTokensUpTo(self, expected_token_type, expected_token):
    return self._GetVarTokensUpTo(expected_token_type, expected_token)[0]

  def _GetVarTokensUpTo(self, expected_token_type, *expected_tokens):
    last_token = self._GetNextToken()
    tokens = []
    while (last_token.token_type != expected_token_type or
           last_token.name not in expected_tokens):
      tokens.append(last_token)
      last_token = self._GetNextToken()
    return tokens, last_token

  # Same as _GetVarTokensUpTo, but skips over '<...>' which could contain an
  # expected token.
  def _GetVarTokensUpToIgnoringTemplates(self, expected_token_type,
                                         *expected_tokens):
    last_token = self._GetNextToken()
    tokens = []
    nesting = 0
    while (nesting > 0 or
           last_token.token_type != expected_token_type or
           last_token.name not in expected_tokens):
      tokens.append(last_token)
      last_token = self._GetNextToken()
      if last_token.name == '<':
        nesting += 1
      elif last_token.name == '>':
        nesting -= 1
    return tokens, last_token

  # TODO(nnorwitz): remove _IgnoreUpTo() it shouldn't be necessary.
  def _IgnoreUpTo(self, token_type, token):
    unused_tokens = self._GetTokensUpTo(token_type, token)

  def _SkipIf0Blocks(self):
    count = 1
    while 1:
      token = self._GetNextToken()
      if token.token_type != tokenize.PREPROCESSOR:
        continue

      name = token.name[1:].lstrip()
      if name.startswith('endif'):
        count -= 1
        if count == 0:
          break
      elif name.startswith('if'):
        count += 1

  def _GetMatchingChar(self, open_paren, close_paren, GetNextToken=None):
    if GetNextToken is None:
      GetNextToken = self._GetNextToken
    # Assumes the current token is open_paren and we will consume
    # and return up to the close_paren.
    count = 1
    token = GetNextToken()
    while 1:
      if token.token_type == tokenize.SYNTAX:
        if token.name == open_paren:
          count += 1
        elif token.name == close_paren:
          count -= 1
          if count == 0:
            break
      yield token
      token = GetNextToken()
    yield token

  def _GetParameters(self):
    return self._GetMatchingChar('(', ')')

  def GetScope(self):
    return self._GetMatchingChar('{', '}')

  def _GetNextToken(self):
    if self.token_queue:
      return self.token_queue.pop()
    try:
      return next(self.tokens)
    except StopIteration:
      return

  def _AddBackToken(self, token):
    if token.whence == tokenize.WHENCE_STREAM:
      token.whence = tokenize.WHENCE_QUEUE
      self.token_queue.insert(0, token)
    else:
      assert token.whence == tokenize.WHENCE_QUEUE, token
      self.token_queue.append(token)

  def _AddBackTokens(self, tokens):
    if tokens:
      if tokens[-1].whence == tokenize.WHENCE_STREAM:
        for token in tokens:
          token.whence = tokenize.WHENCE_QUEUE
        self.token_queue[:0] = reversed(tokens)
      else:
        assert tokens[-1].whence == tokenize.WHENCE_QUEUE, tokens
        self.token_queue.extend(reversed(tokens))

  def GetName(self, seq=None):
    """Returns ([tokens], next_token_info)."""
    GetNextToken = self._GetNextToken
    if seq is not None:
      it = iter(seq)
      GetNextToken = lambda: next(it)
    next_token = GetNextToken()
    tokens = []
    last_token_was_name = False
    while (next_token.token_type == tokenize.NAME or
           (next_token.token_type == tokenize.SYNTAX and
            next_token.name in ('::', '<'))):
      # Two NAMEs in a row means the identifier should terminate.
      # It's probably some sort of variable declaration.
      if last_token_was_name and next_token.token_type == tokenize.NAME:
        break
      last_token_was_name = next_token.token_type == tokenize.NAME
      tokens.append(next_token)
      # Handle templated names.
      if next_token.name == '<':
        tokens.extend(self._GetMatchingChar('<', '>', GetNextToken))
        last_token_was_name = True
      next_token = GetNextToken()
    return tokens, next_token

  def GetMethod(self, modifiers, templated_types):
    return_type_and_name = self._GetTokensUpTo(tokenize.SYNTAX, '(')
    assert len(return_type_and_name) >= 1
    return self._GetMethod(return_type_and_name, modifiers, templated_types,
                           False)

  def _GetMethod(self, return_type_and_name, modifiers, templated_types,
                 get_paren):
    template_portion = None
    if get_paren:
      token = self._GetNextToken()
      assert token.token_type == tokenize.SYNTAX, token
      if token.name == '<':
        # Handle templatized dtors.
        template_portion = [token]
        template_portion.extend(self._GetMatchingChar('<', '>'))
        token = self._GetNextToken()
      assert token.token_type == tokenize.SYNTAX, token
      assert token.name == '(', token

    name = return_type_and_name.pop()
    # Handle templatized ctors.
    if name.name == '>':
      index = 1
      while return_type_and_name[index].name != '<':
        index += 1
      template_portion = return_type_and_name[index:] + [name]
      del return_type_and_name[index:]
      name = return_type_and_name.pop()
    elif name.name == ']':
      rt = return_type_and_name
      assert rt[-1].name == '[', return_type_and_name
      assert rt[-2].name == 'operator', return_type_and_name
      name_seq = return_type_and_name[-2:]
      del return_type_and_name[-2:]
      name = tokenize.Token(tokenize.NAME, 'operator[]',
                            name_seq[0].start, name.end)
      # Get the open paren so _GetParameters() below works.
      unused_open_paren = self._GetNextToken()

    # TODO(nnorwitz): store template_portion.
    return_type = return_type_and_name
    indices = name
    if return_type:
      indices = return_type[0]

    # Force ctor for templatized ctors.
    if name.name == self.in_class and not modifiers:
      modifiers |= FUNCTION_CTOR
    parameters = list(self._GetParameters())
    del parameters[-1]              # Remove trailing ')'.

    # Handling operator() is especially weird.
    if name.name == 'operator' and not parameters:
      token = self._GetNextToken()
      assert token.name == '(', token
      parameters = list(self._GetParameters())
      del parameters[-1]          # Remove trailing ')'.

    token = self._GetNextToken()
    while token.token_type == tokenize.NAME:
      modifier_token = token
      token = self._GetNextToken()
      if modifier_token.name == 'const':
        modifiers |= FUNCTION_CONST
      elif modifier_token.name == '__attribute__':
        # TODO(nnorwitz): handle more __attribute__ details.
        modifiers |= FUNCTION_ATTRIBUTE
        assert token.name == '(', token
        # Consume everything between the (parens).
        unused_tokens = list(self._GetMatchingChar('(', ')'))
        token = self._GetNextToken()
      elif modifier_token.name == 'throw':
        modifiers |= FUNCTION_THROW
        assert token.name == '(', token
        # Consume everything between the (parens).
        unused_tokens = list(self._GetMatchingChar('(', ')'))
        token = self._GetNextToken()
      elif modifier_token.name == 'override':
        modifiers |= FUNCTION_OVERRIDE
      elif modifier_token.name == modifier_token.name.upper():
        # HACK(nnorwitz):  assume that all upper-case names
        # are some macro we aren't expanding.
        modifiers |= FUNCTION_UNKNOWN_ANNOTATION
      else:
        self.HandleError('unexpected token', modifier_token)

    assert token.token_type == tokenize.SYNTAX, token
    # Handle ctor initializers.
    if token.name == ':':
      # TODO(nnorwitz): anything else to handle for initializer list?
      while token.name != ';' and token.name != '{':
        token = self._GetNextToken()

    # Handle pointer to functions that are really data but look
    # like method declarations.
    if token.name == '(':
      if parameters[0].name == '*':
        # name contains the return type.
        name = parameters.pop()
        # parameters contains the name of the data.
        modifiers = [p.name for p in parameters]
        # Already at the ( to open the parameter list.
        function_parameters = list(self._GetMatchingChar('(', ')'))
        del function_parameters[-1]  # Remove trailing ')'.
        # TODO(nnorwitz): store the function_parameters.
        token = self._GetNextToken()
        assert token.token_type == tokenize.SYNTAX, token
        assert token.name == ';', token
        return self._CreateVariable(indices, name.name, indices.name,
                                    modifiers, '', None)
      # At this point, we got something like:
      #  return_type (type::*name_)(params);
      # This is a data member called name_ that is a function pointer.
      # With this code: void (sq_type::*field_)(string&);
      # We get: name=void return_type=[] parameters=sq_type ... field_
      # TODO(nnorwitz): is return_type always empty?
      # TODO(nnorwitz): this isn't even close to being correct.
      # Just put in something so we don't crash and can move on.
      real_name = parameters[-1]
      modifiers = [p.name for p in self._GetParameters()]
      del modifiers[-1]           # Remove trailing ')'.
      return self._CreateVariable(indices, real_name.name, indices.name,
                                  modifiers, '', None)

    if token.name == '{':
      body = list(self.GetScope())
      del body[-1]                # Remove trailing '}'.
    else:
      body = None
      if token.name == '=':
        token = self._GetNextToken()

        if token.name == 'default' or token.name == 'delete':
          # Ignore explicitly defaulted and deleted special members
          # in C++11.
          token = self._GetNextToken()
        else:
          # Handle pure-virtual declarations.
          assert token.token_type == tokenize.CONSTANT, token
          assert token.name == '0', token
          modifiers |= FUNCTION_PURE_VIRTUAL
          token = self._GetNextToken()

      if token.name == '[':
        # TODO(nnorwitz): store tokens and improve parsing.
        # template <typename T, size_t N> char (&ASH(T (&seq)[N]))[N];
        tokens = list(self._GetMatchingChar('[', ']'))
        token = self._GetNextToken()

      assert token.name == ';', (token, return_type_and_name, parameters)

    # Looks like we got a method, not a function.
    if len(return_type) > 2 and return_type[-1].name == '::':
      return_type, in_class = \
          self._GetReturnTypeAndClassName(return_type)
      return Method(indices.start, indices.end, name.name, in_class,
                    return_type, parameters, modifiers, templated_types,
                    body, self.namespace_stack)
    return Function(indices.start, indices.end, name.name, return_type,
                    parameters, modifiers, templated_types, body,
                    self.namespace_stack)

  def _GetReturnTypeAndClassName(self, token_seq):
    # Splitting the return type from the class name in a method
    # can be tricky.  For example, Return::Type::Is::Hard::To::Find().
    # Where is the return type and where is the class name?
    # The heuristic used is to pull the last name as the class name.
    # This includes all the templated type info.
    # TODO(nnorwitz): if there is only One name like in the
    # example above, punt and assume the last bit is the class name.

    # Ignore a :: prefix, if exists so we can find the first real name.
    i = 0
    if token_seq[0].name == '::':
      i = 1
    # Ignore a :: suffix, if exists.
    end = len(token_seq) - 1
    if token_seq[end-1].name == '::':
      end -= 1

    # Make a copy of the sequence so we can append a sentinel
    # value. This is required for GetName will has to have some
    # terminating condition beyond the last name.
    seq_copy = token_seq[i:end]
    seq_copy.append(tokenize.Token(tokenize.SYNTAX, '', 0, 0))
    names = []
    while i < end:
      # Iterate through the sequence parsing out each name.
      new_name, next = self.GetName(seq_copy[i:])
      assert new_name, 'Got empty new_name, next=%s' % next
      # We got a pointer or ref.  Add it to the name.
      if next and next.token_type == tokenize.SYNTAX:
        new_name.append(next)
      names.append(new_name)
      i += len(new_name)

    # Now that we have the names, it's time to undo what we did.

    # Remove the sentinel value.
    names[-1].pop()
    # Flatten the token sequence for the return type.
    return_type = [e for seq in names[:-1] for e in seq]
    # The class name is the last name.
    class_name = names[-1]
    return return_type, class_name

  def handle_bool(self):
    pass

  def handle_char(self):
    pass

  def handle_int(self):
    pass

  def handle_long(self):
    pass

  def handle_short(self):
    pass

  def handle_double(self):
    pass

  def handle_float(self):
    pass

  def handle_void(self):
    pass

  def handle_wchar_t(self):
    pass

  def handle_unsigned(self):
    pass

  def handle_signed(self):
    pass

  def _GetNestedType(self, ctor):
    name = None
    name_tokens, token = self.GetName()
    if name_tokens:
      name = ''.join([t.name for t in name_tokens])

    # Handle forward declarations.
    if token.token_type == tokenize.SYNTAX and token.name == ';':
      return ctor(token.start, token.end, name, None,
                  self.namespace_stack)

    if token.token_type == tokenize.NAME and self._handling_typedef:
      self._AddBackToken(token)
      return ctor(token.start, token.end, name, None,
                  self.namespace_stack)

    # Must be the type declaration.
    fields = list(self._GetMatchingChar('{', '}'))
    del fields[-1]                  # Remove trailing '}'.
    if token.token_type == tokenize.SYNTAX and token.name == '{':
      next = self._GetNextToken()
      new_type = ctor(token.start, token.end, name, fields,
                      self.namespace_stack)
      # A name means this is an anonymous type and the name
      # is the variable declaration.
      if next.token_type != tokenize.NAME:
        return new_type
      name = new_type
      token = next

    # Must be variable declaration using the type prefixed with keyword.
    assert token.token_type == tokenize.NAME, token
    return self._CreateVariable(token, token.name, name, [], '', None)

  def handle_struct(self):
    # Special case the handling typedef/aliasing of structs here.
    # It would be a pain to handle in the class code.
    name_tokens, var_token = self.GetName()
    if name_tokens:
      next_token = self._GetNextToken()
      is_syntax = (var_token.token_type == tokenize.SYNTAX and
                   var_token.name[0] in '*&')
      is_variable = (var_token.token_type == tokenize.NAME and
                     next_token.name == ';')
      variable = var_token
      if is_syntax and not is_variable:
        variable = next_token
        temp = self._GetNextToken()
        if temp.token_type == tokenize.SYNTAX and temp.name == '(':
          # Handle methods declared to return a struct.
          t0 = name_tokens[0]
          struct = tokenize.Token(tokenize.NAME, 'struct',
                                  t0.start-7, t0.start-2)
          type_and_name = [struct]
          type_and_name.extend(name_tokens)
          type_and_name.extend((var_token, next_token))
          return self._GetMethod(type_and_name, 0, None, False)
        assert temp.name == ';', (temp, name_tokens, var_token)
      if is_syntax or (is_variable and not self._handling_typedef):
        modifiers = ['struct']
        type_name = ''.join([t.name for t in name_tokens])
        position = name_tokens[0]
        return self._CreateVariable(position, variable.name, type_name,
                                    modifiers, var_token.name, None)
      name_tokens.extend((var_token, next_token))
      self._AddBackTokens(name_tokens)
    else:
      self._AddBackToken(var_token)
    return self._GetClass(Struct, VISIBILITY_PUBLIC, None)

  def handle_union(self):
    return self._GetNestedType(Union)

  def handle_enum(self):
    return self._GetNestedType(Enum)

  def handle_auto(self):
    # TODO(nnorwitz): warn about using auto?  Probably not since it
    # will be reclaimed and useful for C++0x.
    pass

  def handle_register(self):
    pass

  def handle_const(self):
    pass

  def handle_inline(self):
    pass

  def handle_extern(self):
    pass

  def handle_static(self):
    pass

  def handle_virtual(self):
    # What follows must be a method.
    token = token2 = self._GetNextToken()
    if token.name == 'inline':
      # HACK(nnorwitz): handle inline dtors by ignoring 'inline'.
      token2 = self._GetNextToken()
    if token2.token_type == tokenize.SYNTAX and token2.name == '~':
      return self.GetMethod(FUNCTION_VIRTUAL + FUNCTION_DTOR, None)
    assert token.token_type == tokenize.NAME or token.name == '::', token
    return_type_and_name, _ = self._GetVarTokensUpToIgnoringTemplates(
        tokenize.SYNTAX, '(')  # )
    return_type_and_name.insert(0, token)
    if token2 is not token:
      return_type_and_name.insert(1, token2)
    return self._GetMethod(return_type_and_name, FUNCTION_VIRTUAL,
                           None, False)

  def handle_volatile(self):
    pass

  def handle_mutable(self):
    pass

  def handle_public(self):
    assert self.in_class
    self.visibility = VISIBILITY_PUBLIC

  def handle_protected(self):
    assert self.in_class
    self.visibility = VISIBILITY_PROTECTED

  def handle_private(self):
    assert self.in_class
    self.visibility = VISIBILITY_PRIVATE

  def handle_friend(self):
    tokens = self._GetTokensUpTo(tokenize.SYNTAX, ';')
    assert tokens
    t0 = tokens[0]
    return Friend(t0.start, t0.end, tokens, self.namespace_stack)

  def handle_static_cast(self):
    pass

  def handle_const_cast(self):
    pass

  def handle_dynamic_cast(self):
    pass

  def handle_reinterpret_cast(self):
    pass

  def handle_new(self):
    pass

  def handle_delete(self):
    tokens = self._GetTokensUpTo(tokenize.SYNTAX, ';')
    assert tokens
    return Delete(tokens[0].start, tokens[0].end, tokens)

  def handle_typedef(self):
    token = self._GetNextToken()
    if (token.token_type == tokenize.NAME and
            keywords.IsKeyword(token.name)):
      # Token must be struct/enum/union/class.
      method = getattr(self, 'handle_' + token.name)
      self._handling_typedef = True
      tokens = [method()]
      self._handling_typedef = False
    else:
      tokens = [token]

    # Get the remainder of the typedef up to the semi-colon.
    tokens.extend(self._GetTokensUpTo(tokenize.SYNTAX, ';'))

    # TODO(nnorwitz): clean all this up.
    assert tokens
    name = tokens.pop()
    indices = name
    if tokens:
      indices = tokens[0]
    if not indices:
      indices = token
    if name.name == ')':
      # HACK(nnorwitz): Handle pointers to functions "properly".
      if (len(tokens) >= 4 and
              tokens[1].name == '(' and tokens[2].name == '*'):
        tokens.append(name)
        name = tokens[3]
    elif name.name == ']':
      # HACK(nnorwitz): Handle arrays properly.
      if len(tokens) >= 2:
        tokens.append(name)
        name = tokens[1]
    new_type = tokens
    if tokens and isinstance(tokens[0], tokenize.Token):
      new_type = self.converter.ToType(tokens)[0]
    return Typedef(indices.start, indices.end, name.name,
                   new_type, self.namespace_stack)

  def handle_typeid(self):
    pass  # Not needed yet.

  def handle_typename(self):
    pass  # Not needed yet.

  def _GetTemplatedTypes(self):
    result = collections.OrderedDict()
    tokens = list(self._GetMatchingChar('<', '>'))
    len_tokens = len(tokens) - 1    # Ignore trailing '>'.
    i = 0
    while i < len_tokens:
      key = tokens[i].name
      i += 1
      if keywords.IsKeyword(key) or key == ',':
        continue
      type_name = default = None
      if i < len_tokens:
        i += 1
        if tokens[i-1].name == '=':
          assert i < len_tokens, '%s %s' % (i, tokens)
          default, unused_next_token = self.GetName(tokens[i:])
          i += len(default)
        else:
          if tokens[i-1].name != ',':
            # We got something like: Type variable.
            # Re-adjust the key (variable) and type_name (Type).
            key = tokens[i-1].name
            type_name = tokens[i-2]

      result[key] = (type_name, default)
    return result

  def handle_template(self):
    token = self._GetNextToken()
    assert token.token_type == tokenize.SYNTAX, token
    assert token.name == '<', token
    templated_types = self._GetTemplatedTypes()
    # TODO(nnorwitz): for now, just ignore the template params.
    token = self._GetNextToken()
    if token.token_type == tokenize.NAME:
      if token.name == 'class':
        return self._GetClass(Class, VISIBILITY_PRIVATE, templated_types)
      elif token.name == 'struct':
        return self._GetClass(Struct, VISIBILITY_PUBLIC, templated_types)
      elif token.name == 'friend':
        return self.handle_friend()
    self._AddBackToken(token)
    tokens, last = self._GetVarTokensUpTo(tokenize.SYNTAX, '(', ';')
    tokens.append(last)
    self._AddBackTokens(tokens)
    if last.name == '(':
      return self.GetMethod(FUNCTION_NONE, templated_types)
    # Must be a variable definition.
    return None

  def handle_true(self):
    pass  # Nothing to do.

  def handle_false(self):
    pass  # Nothing to do.

  def handle_asm(self):
    pass  # Not needed yet.

  def handle_class(self):
    return self._GetClass(Class, VISIBILITY_PRIVATE, None)

  def _GetBases(self):
    # Get base classes.
    bases = []
    while 1:
      token = self._GetNextToken()
      assert token.token_type == tokenize.NAME, token
      # TODO(nnorwitz): store kind of inheritance...maybe.
      if token.name not in ('public', 'protected', 'private'):
        # If inheritance type is not specified, it is private.
        # Just put the token back so we can form a name.
        # TODO(nnorwitz): it would be good to warn about this.
        self._AddBackToken(token)
      else:
        # Check for virtual inheritance.
        token = self._GetNextToken()
        if token.name != 'virtual':
          self._AddBackToken(token)
        else:
          # TODO(nnorwitz): store that we got virtual for this base.
          pass
      base, next_token = self.GetName()
      bases_ast = self.converter.ToType(base)
      assert len(bases_ast) == 1, bases_ast
      bases.append(bases_ast[0])
      assert next_token.token_type == tokenize.SYNTAX, next_token
      if next_token.name == '{':
        token = next_token
        break
      # Support multiple inheritance.
      assert next_token.name == ',', next_token
    return bases, token

  def _GetClass(self, class_type, visibility, templated_types):
    class_name = None
    class_token = self._GetNextToken()
    if class_token.token_type != tokenize.NAME:
      assert class_token.token_type == tokenize.SYNTAX, class_token
      token = class_token
    else:
      # Skip any macro (e.g. storage class specifiers) after the
      # 'class' keyword.
      next_token = self._GetNextToken()
      if next_token.token_type == tokenize.NAME:
        self._AddBackToken(next_token)
      else:
        self._AddBackTokens([class_token, next_token])
      name_tokens, token = self.GetName()
      class_name = ''.join([t.name for t in name_tokens])
    bases = None
    if token.token_type == tokenize.SYNTAX:
      if token.name == ';':
        # Forward declaration.
        return class_type(class_token.start, class_token.end,
                          class_name, None, templated_types, None,
                          self.namespace_stack)
      if token.name in '*&':
        # Inline forward declaration.  Could be method or data.
        name_token = self._GetNextToken()
        next_token = self._GetNextToken()
        if next_token.name == ';':
          # Handle data
          modifiers = ['class']
          return self._CreateVariable(class_token, name_token.name,
                                      class_name,
                                      modifiers, token.name, None)
        else:
          # Assume this is a method.
          tokens = (class_token, token, name_token, next_token)
          self._AddBackTokens(tokens)
          return self.GetMethod(FUNCTION_NONE, None)
      if token.name == ':':
        bases, token = self._GetBases()

    body = None
    if token.token_type == tokenize.SYNTAX and token.name == '{':
      assert token.token_type == tokenize.SYNTAX, token
      assert token.name == '{', token

      ast = AstBuilder(self.GetScope(), self.filename, class_name,
                       visibility, self.namespace_stack)
      body = list(ast.Generate())

      if not self._handling_typedef:
        token = self._GetNextToken()
        if token.token_type != tokenize.NAME:
          assert token.token_type == tokenize.SYNTAX, token
          assert token.name == ';', token
        else:
          new_class = class_type(class_token.start, class_token.end,
                                 class_name, bases, None,
                                 body, self.namespace_stack)

          modifiers = []
          return self._CreateVariable(class_token,
                                      token.name, new_class,
                                      modifiers, token.name, None)
    else:
      if not self._handling_typedef:
        self.HandleError('non-typedef token', token)
      self._AddBackToken(token)

    return class_type(class_token.start, class_token.end, class_name,
                      bases, templated_types, body, self.namespace_stack)

  def handle_namespace(self):
    # Support anonymous namespaces.
    name = None
    name_tokens, token = self.GetName()
    if name_tokens:
      name = ''.join([t.name for t in name_tokens])
    self.namespace_stack.append(name)
    assert token.token_type == tokenize.SYNTAX, token
    # Create an internal token that denotes when the namespace is complete.
    internal_token = tokenize.Token(_INTERNAL_TOKEN, _NAMESPACE_POP,
                                    None, None)
    internal_token.whence = token.whence
    if token.name == '=':
      # TODO(nnorwitz): handle aliasing namespaces.
      name, next_token = self.GetName()
      assert next_token.name == ';', next_token
      self._AddBackToken(internal_token)
    else:
      assert token.name == '{', token
      tokens = list(self.GetScope())
      # Replace the trailing } with the internal namespace pop token.
      tokens[-1] = internal_token
      # Handle namespace with nothing in it.
      self._AddBackTokens(tokens)
    return None

  def handle_using(self):
    tokens = self._GetTokensUpTo(tokenize.SYNTAX, ';')
    assert tokens
    return Using(tokens[0].start, tokens[0].end, tokens)

  def handle_explicit(self):
    assert self.in_class
    # Nothing much to do.
    # TODO(nnorwitz): maybe verify the method name == class name.
    # This must be a ctor.
    return self.GetMethod(FUNCTION_CTOR, None)

  def handle_this(self):
    pass  # Nothing to do.

  def handle_operator(self):
    # Pull off the next token(s?) and make that part of the method name.
    pass

  def handle_sizeof(self):
    pass

  def handle_case(self):
    pass

  def handle_switch(self):
    pass

  def handle_default(self):
    token = self._GetNextToken()
    assert token.token_type == tokenize.SYNTAX
    assert token.name == ':'

  def handle_if(self):
    pass

  def handle_else(self):
    pass

  def handle_return(self):
    tokens = self._GetTokensUpTo(tokenize.SYNTAX, ';')
    if not tokens:
      return Return(self.current_token.start, self.current_token.end, None)
    return Return(tokens[0].start, tokens[0].end, tokens)

  def handle_goto(self):
    tokens = self._GetTokensUpTo(tokenize.SYNTAX, ';')
    assert len(tokens) == 1, str(tokens)
    return Goto(tokens[0].start, tokens[0].end, tokens[0].name)

  def handle_try(self):
    pass  # Not needed yet.

  def handle_catch(self):
    pass  # Not needed yet.

  def handle_throw(self):
    pass  # Not needed yet.

  def handle_while(self):
    pass

  def handle_do(self):
    pass

  def handle_for(self):
    pass

  def handle_break(self):
    self._IgnoreUpTo(tokenize.SYNTAX, ';')

  def handle_continue(self):
    self._IgnoreUpTo(tokenize.SYNTAX, ';')


def BuilderFromSource(source, filename):
  """Utility method that returns an AstBuilder from source code.

    Args:
      source: 'C++ source code'
      filename: 'file1'

    Returns:
      AstBuilder
    """
  return AstBuilder(tokenize.GetTokens(source), filename)


def PrintIndentifiers(filename, should_print):
  """Prints all identifiers for a C++ source file.

    Args:
      filename: 'file1'
      should_print: predicate with signature: bool Function(token)
    """
  source = utils.ReadFile(filename, False)
  if source is None:
    sys.stderr.write('Unable to find: %s\n' % filename)
    return

  #print('Processing %s' % actual_filename)
  builder = BuilderFromSource(source, filename)
  try:
    for node in builder.Generate():
      if should_print(node):
        print(node.name)
  except KeyboardInterrupt:
    return
  except:
    pass


def PrintAllIndentifiers(filenames, should_print):
  """Prints all identifiers for each C++ source file in filenames.

    Args:
      filenames: ['file1', 'file2', ...]
      should_print: predicate with signature: bool Function(token)
    """
  for path in filenames:
    PrintIndentifiers(path, should_print)


def main(argv):
  for filename in argv[1:]:
    source = utils.ReadFile(filename)
    if source is None:
      continue

    print('Processing %s' % filename)
    builder = BuilderFromSource(source, filename)
    try:
      entire_ast = filter(None, builder.Generate())
    except KeyboardInterrupt:
      return
    except:
      # Already printed a warning, print the traceback and continue.
      traceback.print_exc()
    else:
      if utils.DEBUG:
        for ast in entire_ast:
          print(ast)


if __name__ == '__main__':
  main(sys.argv)



from sealy import (BFVEncoder, BfvEncryptionParametersBuilder,
                   CoefficientModulus, Context, DegreeType, PlainModulus,
                   SecurityLevel)


def test_can_create_and_drop_bfv_encoder():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    encoder = BFVEncoder(ctx)
    del encoder


def test_can_get_slots_bfv_encoder():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    encoder = BFVEncoder(ctx)

    assert encoder.get_slot_count() == 8192


def test_can_get_encode_and_decode_unsigned():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    encoder = BFVEncoder(ctx)

    data = [i for i in range(encoder.get_slot_count())]

    plaintext = encoder.encode(data)
    data_2 = encoder.decode(plaintext)

    assert data == data_2


def test_can_get_encode_and_decode_signed():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    encoder = BFVEncoder(ctx)

    data = [i for i in range(encoder.get_slot_count())]

    plaintext = encoder.encode(data)
    data_2 = encoder.decode(plaintext)

    assert data == data_2


def test_scalar_encoder_can_encode_decode_signed():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    encoder = BFVEncoder(ctx)

    p = encoder.encode([-15])

    assert encoder.decode(p)[0] == -15


def test_scalar_encoder_can_encode_decode_unsigned():
    params = (
        BfvEncryptionParametersBuilder()
        .with_poly_modulus_degree(DegreeType(8192))
        .with_coefficient_modulus(
            CoefficientModulus.create(DegreeType(8192), [50, 30, 30, 50, 50])
        )
        .with_plain_modulus(PlainModulus.batching(DegreeType(8192), 20))
        .build()
    )

    ctx = Context(params, False, SecurityLevel(128))
    encoder = BFVEncoder(ctx)

    p = encoder.encode([42])

    assert encoder.decode(p)[0] == 42



from typing import List, Tuple

class MemoryPool:
    """
    Represents the memory pool used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new memory pool.
        """
        ...

class SchemeType:
    """
    Represents a scheme type used in encryption parameters.
    """

    def __init__(self, val: int) -> None:
        """
        Initialize a new SchemeType with a given value.

        :param val: The value representing the scheme type.
        """
        ...

    @classmethod
    def bfv(cls) -> "SchemeType":
        """
        Create a BFV scheme type.

        :return: A SchemeType instance for BFV.
        """
        ...

    @classmethod
    def ckks(cls) -> "SchemeType":
        """
        Create a CKKS scheme type.

        :return: A SchemeType instance for CKKS.
        """
        ...

class CoefficientModulus:
    """
    Represents the coefficient modulus used in encryption parameters.
    """

    @classmethod
    def create(
        cls, degree: "DegreeType", bit_sizes: List[int]
    ) -> List["Modulus"]:
        """
        Initialize a new coefficient modulus with a given degree and bit sizes.

        :param degree: The polynomial degree.
        :param bit_sizes: A list of bit sizes for the moduli.
        :return: A list of moduli.
        """
        ...

    @classmethod
    def ckks(
        cls, degree: "DegreeType", bit_sizes: List[int]
    ) -> List["Modulus"]:
        """
        Create a CKKS coefficient modulus.

        :param degree: The polynomial degree.
        :param bit_sizes: A list of bit sizes for the moduli.
        :return: A list of moduli.
        """
        ...

    @classmethod
    def bfv(
        cls, degree: "DegreeType", security_level: "SecurityLevel"
    ) -> List["Modulus"]:
        """
        Create a BFV coefficient modulus.

        :param degree: The polynomial degree.
        :param security_level: The security level.
        :return: A list of moduli.
        """
        ...

    @classmethod
    def max_bit_count(
        cls, degree: "DegreeType", security_level: "SecurityLevel"
    ) -> int:
        """
        Get the maximum bit count for a given degree and security level.

        :param degree: The polynomial degree.
        :param security_level: The security level.
        :return: The maximum bit count.
        """
        ...

class PlainModulus:
    """
    Represents the plain modulus used in encryption parameters.
    """

    @classmethod
    def batching(cls, degree: "DegreeType", bit_size: int) -> "Modulus":
        """
        Create a plain modulus for batching.

        :param degree: The polynomial degree.
        :param bit_size: The bit size for the modulus.
        :return: The plain modulus.
        """
        ...

    @classmethod
    def raw(cls, val: int) -> "Modulus":
        """
        Create a plain modulus from a raw value.

        :param val: The raw value for the modulus.
        :return: The plain modulus.
        """
        ...

class Modulus:
    """
    Represents a modulus used in encryption parameters.
    """

    def __init__(self, value: int) -> None:
        """
        Initialize a new modulus with a given value.

        :param value: The value of the modulus.
        """
        ...

    def get_value(self) -> int:
        """
        Get the value of the modulus.

        :return: The value of the modulus.
        """
        ...

class DegreeType:
    """
    Represents a degree type used in encryption parameters.
    """

    def __init__(self, degree: int) -> None:
        """
        Initialize a new degree type with a given value.

        :param degree: The degree value.
        """
        ...

class SecurityLevel:
    """
    Represents a security level used in encryption parameters.
    """

    def __init__(self, value: int) -> None:
        """
        Initialize a new security level with a given value.

        :param value: The security level value.
        """
        ...

    @classmethod
    def default(cls) -> "SecurityLevel":
        """
        Get the default security level.

        :return: The default security level.
        """
        ...

    def get_value(self) -> int:
        """
        Get the value of the security level.

        :return: The value of the security level.
        """
        ...

class EncryptionParameters:
    """
    Represents the encryption parameters for the scheme.
    """

    def __init__(self, scheme: SchemeType) -> None:
        """
        Initialize encryption parameters with a given scheme type.

        :param scheme: The scheme type for the encryption parameters.
        """
        ...

    @classmethod
    def get_block_size(cls) -> int:
        """
        Get the block size of the encryption parameters.

        :return: The block size.
        """
        ...

    def get_poly_modulus_degree(self) -> int:
        """
        Get the polynomial modulus degree of the encryption parameters.

        :return: The polynomial modulus degree.
        """
        ...

    def get_scheme(self) -> SchemeType:
        """
        Get the scheme type of the encryption parameters.

        :return: The scheme type.
        """
        ...

    def get_plain_modulus(self) -> "Modulus":
        """
        Get the plain modulus of the encryption parameters.

        :return: The plain modulus.
        """
        ...

    def get_coefficient_modulus(self) -> List["Modulus"]:
        """
        Get the coefficient modulus of the encryption parameters.

        :return: A list of coefficient moduli.
        """
        ...

    def get_parms_id(self) -> int:
        """
        Get the parameters ID of the encryption parameters.

        :return: The parameters ID.
        """
        ...

    def set_coefficient_modulus(self, modulus: List["Modulus"]) -> None:
        """
        Set the coefficient modulus of the encryption parameters.

        :param modulus: A list of coefficient moduli to set.
        """
        ...

    def set_poly_modulus_degree(self, degree: "DegreeType") -> None:
        """
        Set the polynomial modulus degree of the encryption parameters.

        :param degree: The polynomial modulus degree to set.
        """
        ...

    def set_plain_modulus(self, modulus: "Modulus") -> None:
        """
        Set the plain modulus of the encryption parameters.

        :param modulus: The plain modulus to set.
        """
        ...

    def set_plain_modulus_constant(self, modulus: int) -> None:
        """
        Set the plain modulus of the encryption parameters using a constant value.

        :param modulus: The constant value for the plain modulus.
        """
        ...

class ContextData:
    """
    Represents the context data used in encryption parameters.
    """

    def get_encryption_parameters(self) -> "EncryptionParameters":
        """
        Get the encryption parameters from the context data.

        :return: The encryption parameters.
        """
        ...

    def get_total_coeff_modulus_bit_count(self) -> int:
        """
        Get the total coefficient modulus bit count from the context data.

        :return: The total coefficient modulus bit count.
        """
        ...

class Context:
    """
    Represents the context used in encryption parameters.
    """

    def __init__(
        self,
        params: "EncryptionParameters",
        expand_mod_chain: bool,
        security_level: "SecurityLevel",
    ) -> None:
        """
        Initialize a new context with given parameters.

        :param params: The encryption parameters.
        :param expand_mod_chain: A flag indicating whether to expand the modulus chain.
        :param security_level: The security level.
        """
        ...

    def get_key_parms_id(self) -> List[int]:
        """
        Get the key parameters ID from the context.

        :return: The key parameters ID.
        """
        ...

    def get_last_parms_id(self) -> int:
        """
        Get the last parameters ID from the context.

        :return: The last parameters ID.
        """
        ...

    def get_first_parms_id(self) -> int:
        """
        Get the first parameters ID from the context.

        :return: The first parameters ID.
        """
        ...

    def get_context_data(self) -> "ContextData":
        """
        Get the context data from the context.

        :return: The context data.
        """
        ...

class PublicKey:
    """
    Represents the public key used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new public key.
        """
        ...

    def as_bytes(self) -> bytes:
        """
        Convert the public key to a list of bytes.
        """
        ...

    @classmethod
    def from_bytes(cls, context: "Context", data: bytes) -> "PublicKey":
        """
        Load the public key from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

class SecretKey:
    """
    Represents the secret key used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new secret key.
        """
        ...

    def as_bytes(self) -> bytes:
        """
        Convert the secret key to a list of bytes.
        """
        ...

    @classmethod
    def from_bytes(cls, context: "Context", data: bytes) -> "SecretKey":
        """
        Load the secret key from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

class RelinearizationKey:
    """
    Represents the relinearization keys used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new relinearization keys.
        """
        ...

    def as_bytes(self) -> bytes:
        """
        Convert the relinearization keys to a list of bytes.
        """
        ...

    @classmethod
    def from_bytes(
        cls, context: "Context", data: bytes
    ) -> "RelinearizationKey":
        """
        Load the relinearization keys from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

class GaloisKey:
    """
    Represents the Galois keys used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new Galois keys.
        """
        ...

    def as_bytes(self) -> bytes:
        """
        Convert the Galois keys to a list of bytes.
        """
        ...

    @classmethod
    def from_bytes(cls, context: "Context", data: bytes) -> "GaloisKey":
        """
        Load the Galois keys from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

class KeyGenerator:
    """
    Represents the key generator used in encryption parameters.
    """

    def __init__(self, context: "Context") -> None:
        """
        Initialize a new key generator with a given context.

        :param context: The context for the key generator.
        """
        ...

    @classmethod
    def from_secret_key(
        cls, context: "Context", secret_key: "SecretKey"
    ) -> "KeyGenerator":
        """
        Generate a public key.

        :return: The public key.
        """
        ...

    def secret_key(self) -> "SecretKey":
        """
        Copies the secret key.

        :return: The secret key.
        """
        ...

    def create_public_key(self) -> "PublicKey":
        """
        Generate a public key.

        :return: The public key.
        """
        ...

    def create_galois_key(self) -> "GaloisKey":
        """
        Generate Galois keys.

        :return: The Galois keys.
        """
        ...

    def create_relinearization_key(self) -> "RelinearizationKey":
        """
        Generate relinearization keys.

        :return: The relinearization keys.
        """
        ...

class Plaintext:
    """
    Represents the plaintext used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new plaintext.
        """
        ...

    @classmethod
    def with_pool(cls, pool: "MemoryPool") -> "Plaintext":
        """
        Initialize a new plaintext with a given memory pool.

        :param pool: The memory pool.
        """
        ...

    def as_bytes(self) -> bytes:
        """
        Convert the plaintext to a list of bytes.
        """
        ...

    @classmethod
    def from_bytes(cls, context: "Context", data: bytes) -> "Plaintext":
        """
        Load the plaintext from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

    @classmethod
    def from_hex_string(cls, hex_string: str) -> "Plaintext":
        """
        Load the plaintext from a hex string.

        :param hex_string: The hex string to load.
        """
        ...

    def get_coefficient(self, index: int) -> int:
        """
        Get the coefficient at a given index.

        :param index: The index of the coefficient.
        :return: The coefficient at the given index.
        """
        ...

    def set_coefficient(self, index: int, value: int) -> None:
        """
        Set the coefficient at a given index.

        :param index: The index of the coefficient.
        :param value: The value to set.
        """
        ...

    def resize(self, size: int) -> None:
        """
        Resize the plaintext to a given size.

        :param size: The size to resize to.
        """
        ...

    def size(self) -> int:
        """
        Get the size of the plaintext.

        :return: The size of the plaintext.
        """
        ...

    def is_ntt_form(self) -> bool:
        """
        Check if the plaintext is in NTT form.

        :return: True if the plaintext is in NTT form, False otherwise.
        """
        ...

class Ciphertext:
    """
    Represents the ciphertext used in encryption parameters.
    """

    def __init__(self) -> None:
        """
        Initialize a new ciphertext.
        """
        ...

    def as_bytes(self) -> bytes:
        """
        Convert the ciphertext to a list of bytes.
        """
        ...

    @classmethod
    def from_bytes(cls, context: "Context", data: bytes) -> "Ciphertext":
        """
        Load the ciphertext from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

    def get_num_polynomials(self) -> int:
        """
        Get the number of polynomials in the ciphertext.

        :return: The number of polynomials.
        """
        ...

    def get_coefficient(self, poly_index: int, coeff_index: int) -> int:
        """
        Get the coefficient at a given polynomial and coefficient index.

        :param poly_index: The polynomial index.
        :param coeff_index: The coefficient index.
        :return: The coefficient at the given polynomial and coefficient index.
        """
        ...

    def get_coeff_modulus_size(self) -> int:
        """
        Get the coefficient modulus size of the ciphertext.

        :return: The coefficient modulus size.
        """
        ...

    def is_ntt_form(self) -> bool:
        """
        Check if the ciphertext is in NTT form.

        :return: True if the ciphertext is in NTT form, False otherwise.
        """
        ...

from typing import List, Optional

class PolynomialArray:
    def __init__(self) -> None:
        """Creates a new empty polynomial array. Use an encoder to populate with a value."""
        ...

    @classmethod
    def from_ciphertext(
        cls, context: "Context", ciphertext: "Ciphertext"
    ) -> "PolynomialArray":
        """Creates a polynomial array from a reference to a ciphertext."""
        ...

    @classmethod
    def from_public_key(
        cls, context: "Context", public_key: "PublicKey"
    ) -> "PolynomialArray":
        """Creates a polynomial array from a reference to a public key."""
        ...

    @classmethod
    def from_secret_key(
        cls, context: "Context", secret_key: "SecretKey"
    ) -> "PolynomialArray":
        """Creates a polynomial array from a reference to a secret key."""
        ...

    def is_reserved(self) -> bool:
        """Has the array data been loaded? When an array is created, it initially
        has no data. Once data is loaded this is true. Additionally data can only
        be loaded once."""
        ...

    def is_rns(self) -> bool:
        """Is the array in RNS form (true)."""
        ...

    def is_multiprecision(self) -> bool:
        """Is the array in RNS form (true)."""
        ...

    def to_rns(self) -> None:
        """Converts the polynomial array into the RNS format regardless of its
        current format."""
        ...

    def to_multiprecision(self) -> None:
        """Converts the polynomial array into the multiprecision format regardless
        of its current format."""
        ...

    def as_multiprecision_bytes(self) -> bytes:
        """This will be in coefficient order; all the limbs with a given coefficient
        are stored together in least significant order.

        The number of limbs equals the number of moduli in the coefficient
        modulus."""
        ...

    def as_rns_bytes(self) -> bytes:
        """This will be in modulus order; all the values associated with a given
        moduli are stored together.

        The number of limbs equals the number of moduli in the coefficient
        modulus."""
        ...

    def as_ints(self) -> List[int]:
        """Returns the array as a list of integers."""
        ...

    def get_num_polynomials(self) -> int:
        """Returns the number of polynomials stored in the `PolynomialArray`."""
        ...

    def get_poly_modulus_degree(self) -> int:
        """Returns the number of coefficients in each polynomial in the `PolynomialArray`."""
        ...

    def get_coeff_modulus_size(self) -> int:
        """Returns how many moduli are in the coefficient modulus set."""
        ...

    def drop_modulus(self) -> "PolynomialArray":
        """Reduces the polynomial array by dropping the last modulus in the modulus
        set."""
        ...

class BFVEncoder:
    def __init__(self, ctx: "Context") -> None:
        """
        Creates a BFVEncoder.

        Parameters:
        ctx (Context): The context to use for the encoder.
        """
        ...

    def get_slot_count(self) -> int:
        """
        Returns the number of "Batched" slots this encoder produces.

        Returns:
        int: The number of slots.
        """
        ...

    def encode(self, data: List[int]) -> "Plaintext":
        """
        Encodes the given data into a plaintext.

        Parameters:
        data (List[int]): The data to encode.

        Returns:
        Plaintext: The encoded plaintext.
        """
        ...

    def decode(self, plaintext: "Plaintext") -> List[int]:
        """
        Decodes the given plaintext into data.

        Parameters:
        plaintext (Plaintext): The plaintext to decode.

        Returns:
        List[int]: The decoded data.
        """
        ...

class BFVDecimalEncoder:
    def __init__(self, ctx: "Context", base: int) -> None:
        """
        Creates a BFVDecimalEncoder.

        Parameters:
        ctx (PyContext): The context to use for the encoder.
        base (int): The base to use for encoding.
        """
        ...

    def get_slot_count(self) -> int:
        """
        Returns the number of "Batched" slots this encoder produces.

        Returns:
        int: The number of slots.
        """
        ...

    def encode(self, data: List[float]) -> "Plaintext":
        """
        Encodes the given data into a plaintext.

        Parameters:
        data (List[float]): The data to encode.

        Returns:
        PyPlaintext: The encoded plaintext.
        """
        ...

    def decode(self, plaintext: "Plaintext") -> List[float]:
        """
        Decodes the given plaintext into data.

        Parameters:
        plaintext (PyPlaintext): The plaintext to decode.

        Returns:
        List[float]: The decoded data.
        """
        ...

class CKKSEncoder:
    def __init__(self, ctx: "Context", scale: float) -> None:
        """
        Creates a CKKSEncoder.

        Parameters:
        ctx (Context): The context to use for the encoder.
        scale (float): The scale to use for encoding.
        """
        ...

    def get_slot_count(self) -> int:
        """
        Returns the number of "Batched" slots this encoder produces.

        Returns:
        int: The number of slots.
        """
        ...

    def encode(self, data: List[float]) -> "Plaintext":
        """
        Encodes the given data into a plaintext.

        Parameters:
        data (List[float]): The data to encode.

        Returns:
        Plaintext: The encoded plaintext.
        """
        ...

    def decode(self, plaintext: "Plaintext") -> List[float]:
        """
        Decodes the given plaintext into data.

        Parameters:
        plaintext (Plaintext): The plaintext to decode.

        Returns:
        List[float]: The decoded data.
        """
        ...

class AsymmetricComponents:
    """
    Contains the asymmetric components of the encryption scheme.
    """

    def __init__(
        self,
        u: "PolynomialArray",
        e: "PolynomialArray",
        r: "Plaintext",
    ) -> None:
        """
        Initializes the asymmetric components.

        Parameters:
        u (PolynomialArray): The first component.
        e (PolynomialArray): The second component.
        r (Plaintext): The plaintext component.
        """
        ...

    def get_u(self) -> "PolynomialArray":
        """
        Gets the first component.

        Returns:
        PolynomialArray: The first component.
        """
        ...

    def get_e(self) -> "PolynomialArray":
        """
        Gets the second component.

        Returns:
        PolynomialArray: The second component.
        """
        ...

    def get_r(self) -> "Plaintext":
        """
        Gets the plaintext component.

        Returns:
        Plaintext: The plaintext component.
        """
        ...

class Encryptor:
    """
    Encrypts plaintext data into ciphertext.
    """

    def __init__(self, ctx: "Context", public_key: "PublicKey") -> None:
        """
        Initializes the encryptor.

        Parameters:
        ctx (Context): The context to use.
        public_key (PublicKey): The public key to use.
        """
        ...

    def encrypt(self, plaintext: "Plaintext") -> "Ciphertext":
        """
        Encrypts the given plaintext.

        Parameters:
        plaintext (Plaintext): The plaintext to encrypt.

        Returns:
        Ciphertext: The encrypted ciphertext.
        """
        ...

    def encrypt_return_components(
        self, plaintext: "Plaintext"
    ) -> Tuple["Ciphertext", "AsymmetricComponents"]:
        """
        Encrypts the given plaintext and returns the components.

        Parameters:
        plaintext (Plaintext): The plaintext to encrypt.

        Returns:
        AsymmetricComponents: The asymmetric components of the encryption.
        """
        ...

class Decryptor:
    """
    Decrypts ciphertext data into plaintext.
    """

    def __init__(self, ctx: "Context", secret_key: "SecretKey") -> None:
        """
        Initializes the decryptor.

        Parameters:
        ctx (Context): The context to use.
        secret_key (SecretKey): The secret key to use.
        """
        ...

    def decrypt(self, ciphertext: "Ciphertext") -> "Plaintext":
        """
        Decrypts the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to decrypt.

        Returns:
        Plaintext: The decrypted plaintext.
        """
        ...

    def invariant_noise_budget(self, ciphertext: "Ciphertext") -> int:
        """
        Returns the invariant noise budget of the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to analyze.

        Returns:
        int: The invariant noise budget.
        """
        ...

    def invariant_noise(self, ciphertext: "Ciphertext") -> float:
        """
        Returns the invariant noise of the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to analyze.

        Returns:
        float: The invariant noise.
        """
        ...

class BFVEvaluator:
    """
    Evaluates operations on BFV ciphertexts.
    """

    def __init__(self, ctx: "Context") -> None:
        """
        Initializes the evaluator.

        Parameters:
        ctx (Context): The context to use.
        """
        ...

    def negate(self, ciphertext: "Ciphertext") -> "Ciphertext":
        """
        Negates the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to negate.

        Returns:
        Ciphertext: The negated ciphertext.
        """
        ...

    def add(self, a: "Ciphertext", b: "Ciphertext") -> "Ciphertext":
        """
        Adds two ciphertexts.

        Parameters:
        a (Ciphertext): The first ciphertext.
        b (Ciphertext): The second ciphertext.

        Returns:
        Ciphertext: The sum of the ciphertexts.
        """
        ...

    def add_many(self, ciphertexts: List["Ciphertext"]) -> "Ciphertext":
        """
        Adds multiple ciphertexts.

        Parameters:
        ciphertexts (List[Ciphertext]): The ciphertexts to add.

        Returns:
        Ciphertext: The sum of the ciphertexts.
        """
        ...

    def multiply(self, a: "Ciphertext", b: "Ciphertext") -> "Ciphertext":
        """
        Multiplies two ciphertexts.

        Parameters:
        a (Ciphertext): The first ciphertext.
        b (Ciphertext): The second ciphertext.

        Returns:
        Ciphertext: The product of the ciphertexts.
        """
        ...

    def multiply_many(
        self, ciphertexts: List["Ciphertext"], relin_key: "RelinearizationKey"
    ) -> "Ciphertext":
        """
        Multiplies multiple ciphertexts.

        Parameters:
        ciphertexts (List[Ciphertext]): The ciphertexts to multiply.
        relin_key (RelinearizationKey): The relinearization key to use.

        Returns:
        Ciphertext: The product of the ciphertexts.
        """
        ...

    def multiply_plain(self, a: "Ciphertext", b: "Plaintext") -> "Ciphertext":
        """
        Multiplies a ciphertext by a plaintext.

        Parameters:
        a (Ciphertext): The ciphertext.
        b (Plaintext): The plaintext.

        Returns:
        Ciphertext: The product of the ciphertext and plaintext.
        """
        ...

    def sub(self, a: "Ciphertext", b: "Ciphertext") -> "Ciphertext":
        """
        Subtracts two ciphertexts.

        Parameters:
        a (Ciphertext): The first ciphertext.
        b (Ciphertext): The second ciphertext.

        Returns:
        Ciphertext: The difference of the ciphertexts.
        """
        ...

    def sub_plain(self, a: "Ciphertext", b: "Plaintext") -> "Ciphertext":
        """
        Subtracts a plaintext from a ciphertext.

        Parameters:
        a (Ciphertext): The ciphertext.
        b (Plaintext): The plaintext.

        Returns:
        Ciphertext: The difference of the ciphertext and plaintext.
        """
        ...

    def relinearize(self, ciphertext: "Ciphertext") -> "Ciphertext":
        """
        Relinearizes the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to relinearize.

        Returns:
        Ciphertext: The relinearized ciphertext.
        """
        ...

class CKKSEvaluator:
    """
    Evaluates operations on BFV ciphertexts.
    """

    def __init__(self, ctx: "Context") -> None:
        """
        Initializes the evaluator.

        Parameters:
        ctx (Context): The context to use.
        """
        ...

    def negate(self, ciphertext: "Ciphertext") -> "Ciphertext":
        """
        Negates the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to negate.

        Returns:
        Ciphertext: The negated ciphertext.
        """
        ...

    def add(self, a: "Ciphertext", b: "Ciphertext") -> "Ciphertext":
        """
        Adds two ciphertexts.

        Parameters:
        a (Ciphertext): The first ciphertext.
        b (Ciphertext): The second ciphertext.

        Returns:
        Ciphertext: The sum of the ciphertexts.
        """
        ...

    def add_many(self, ciphertexts: List["Ciphertext"]) -> "Ciphertext":
        """
        Adds multiple ciphertexts.

        Parameters:
        ciphertexts (List[Ciphertext]): The ciphertexts to add.

        Returns:
        Ciphertext: The sum of the ciphertexts.
        """
        ...

    def multiply(self, a: "Ciphertext", b: "Ciphertext") -> "Ciphertext":
        """
        Multiplies two ciphertexts.

        Parameters:
        a (Ciphertext): The first ciphertext.
        b (Ciphertext): The second ciphertext.

        Returns:
        Ciphertext: The product of the ciphertexts.
        """
        ...

    def multiply_many(
        self, ciphertexts: List["Ciphertext"], relin_key: "RelinearizationKey"
    ) -> "Ciphertext":
        """
        Multiplies multiple ciphertexts.

        Parameters:
        ciphertexts (List[Ciphertext]): The ciphertexts to multiply.
        relin_key (RelinearizationKey): The relinearization key to use.

        Returns:
        Ciphertext: The product of the ciphertexts.
        """
        ...

    def multiply_plain(self, a: "Ciphertext", b: "Plaintext") -> "Ciphertext":
        """
        Multiplies a ciphertext by a plaintext.

        Parameters:
        a (Ciphertext): The ciphertext.
        b (Plaintext): The plaintext.

        Returns:
        Ciphertext: The product of the ciphertext and plaintext.
        """
        ...

    def sub(self, a: "Ciphertext", b: "Ciphertext") -> "Ciphertext":
        """
        Subtracts two ciphertexts.

        Parameters:
        a (Ciphertext): The first ciphertext.
        b (Ciphertext): The second ciphertext.

        Returns:
        Ciphertext: The difference of the ciphertexts.
        """
        ...

    def sub_plain(self, a: "Ciphertext", b: "Plaintext") -> "Ciphertext":
        """
        Subtracts a plaintext from a ciphertext.

        Parameters:
        a (Ciphertext): The ciphertext.
        b (Plaintext): The plaintext.

        Returns:
        Ciphertext: The difference of the ciphertext and plaintext.
        """
        ...

    def relinearize(self, ciphertext: "Ciphertext") -> "Ciphertext":
        """
        Relinearizes the given ciphertext.

        Parameters:
        ciphertext (Ciphertext): The ciphertext to relinearize.

        Returns:
        Ciphertext: The relinearized ciphertext.
        """
        ...

class PlaintextBatchArray:
    """
    Stores a batch of plaintexts.
    """

    def __init__(self, plaintexts: List["Plaintext"]) -> None:
        """
        Initializes the plaintext batch array.

        Parameters:
        plaintexts (List[Plaintext]): The plaintexts to store.
        """
        ...

class CiphertextBatchArray:
    """
    Stores a batch of ciphertexts.
    """

    def __init__(self, ciphertexts: List["Ciphertext"]) -> None:
        """
        Initializes the ciphertext batch array.

        Parameters:
        ciphertexts (List[Ciphertext]): The ciphertexts to store.
        """
        ...

    def as_batched_bytes(self) -> List[bytes]:
        """
        Convert the ciphertexts to a list of bytes.
        """
        ...

    @classmethod
    def from_batched_bytes(
        cls, context: "Context", data: List[bytes]
    ) -> "CiphertextBatchArray":
        """
        Load the ciphertexts from a list of bytes.

        :param data: The list of bytes to load.
        """
        ...

class BatchEncryptor:
    """
    Encrypts a batch of plaintexts into a batch of ciphertexts.
    """

    def __init__(self, ctx: "Context", public_key: "PublicKey") -> None:
        """
        Initializes the batch encryptor.

        Parameters:
        ctx (Context): The context to use.
        public_key (PublicKey): The public key to use.
        """
        ...

    def encrypt(
        self, plaintexts: "PlaintextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Encrypts the given batch of plaintexts.

        Parameters:
        plaintexts (PlaintextBatchArray): The plaintexts to encrypt.

        Returns:
        CiphertextBatchArray: The encrypted ciphertexts.
        """
        ...

class BatchDecryptor:
    """
    Decrypts a batch of ciphertexts into a batch of plaintexts.
    """

    def __init__(self, ctx: "Context", secret_key: "SecretKey") -> None:
        """
        Initializes the batch decryptor.

        Parameters:
        ctx (Context): The context to use.
        secret_key (SecretKey): The secret key to use.
        """
        ...

    def decrypt(
        self, ciphertexts: "CiphertextBatchArray"
    ) -> "PlaintextBatchArray":
        """
        Decrypts the given batch of ciphertexts.

        Parameters:
        ciphertexts (CiphertextBatchArray): The ciphertexts to decrypt.

        Returns:
        PlaintextBatchArray: The decrypted plaintexts.
        """
        ...

class CKKSBatchEncoder:
    """
    Encodes a batch of floating-point values into a batch of plaintexts.
    """

    def __init__(self, ctx: "Context", scale: float) -> None:
        """
        Initializes the batch encoder.

        Parameters:
        ctx (Context): The context to use.
        scale (float): The scale to use for encoding.
        """
        ...

    def encode(self, values: List[float]) -> "PlaintextBatchArray":
        """
        Encodes the given batch of floating-point values.

        Parameters:
        values (List[float]): The values to encode.

        Returns:
        PlaintextBatchArray: The encoded plaintexts.
        """
        ...

    def decode(self, plaintexts: "PlaintextBatchArray") -> List[float]:
        """
        Decodes the given batch of plaintexts.

        Parameters:
        plaintexts (PlaintextBatchArray): The plaintexts to decode.

        Returns:
        List[float]: The decoded values.
        """
        ...

class CKKSBatchEvaluator:
    """
    Evaluates operations on CKKS ciphertexts.
    """

    def __init__(self, ctx: "Context") -> None:
        """
        Initializes the evaluator.

        Parameters:
        ctx (Context): The context to use.
        """
        ...

    def negate(self, a: "CiphertextBatchArray") -> "CiphertextBatchArray":
        """
        Negates the given batch of ciphertexts.

        Parameters:
        a (CiphertextBatchArray): The ciphertexts to negate.

        Returns:
        CiphertextBatchArray: The negated ciphertexts.
        """
        ...

    def add(
        self, a: "CiphertextBatchArray", b: "CiphertextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Adds two batches of ciphertexts.

        Parameters:
        a (CiphertextBatchArray): The first batch of ciphertexts.
        b (CiphertextBatchArray): The second batch of ciphertexts.

        Returns:
        CiphertextBatchArray: The sum of the ciphertexts.
        """
        ...

    def add_many(
        self, ciphertexts: List["CiphertextBatchArray"]
    ) -> "CiphertextBatchArray":
        """
        Adds multiple batches of ciphertexts.

        Parameters:
        ciphertexts (List[CiphertextBatchArray]): The batches of ciphertexts to add.

        Returns:
        CiphertextBatchArray: The sum of the ciphertexts.
        """
        ...

    def multiply(
        self, a: "CiphertextBatchArray", b: "CiphertextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Multiplies two batches of ciphertexts.

        Parameters:
        a (CiphertextBatchArray): The first batch of ciphertexts.
        b (CiphertextBatchArray): The second batch of ciphertexts.

        Returns:
        CiphertextBatchArray: The product of the ciphertexts.
        """
        ...

    def multiply_many(
        self,
        ciphertexts: List["CiphertextBatchArray"],
        relin_key: "RelinearizationKey",
    ) -> "CiphertextBatchArray":
        """
        Multiplies multiple batches of ciphertexts.

        Parameters:
        ciphertexts (List[CiphertextBatchArray]): The batches of ciphertexts to multiply.
        relin_key (RelinearizationKey): The relinearization key to use.

        Returns:
        CiphertextBatchArray: The product of the ciphertexts.
        """
        ...

    def multiply_plain(
        self, a: "CiphertextBatchArray", b: "PlaintextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Multiplies a batch of ciphertexts by a plaintext.

        Parameters:
        a (CiphertextBatchArray): The batch of ciphertexts.
        b (PlaintextBatchArray): The batch of plaintexts.

        Returns:
        CiphertextBatchArray: The product of the ciphertexts and plaintext.
        """
        ...

    def sub(
        self, a: "CiphertextBatchArray", b: "CiphertextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Subtracts two batches of ciphertexts.

        Parameters:
        a (CiphertextBatchArray): The first batch of ciphertexts.
        b (CiphertextBatchArray): The second batch of ciphertexts.

        Returns:
        CiphertextBatchArray: The difference of the ciphertexts.
        """
        ...

    def sub_plain(
        self, a: "CiphertextBatchArray", b: "PlaintextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Subtracts a batch of plaintexts from a batch of ciphertexts.

        Parameters:
        a (CiphertextBatchArray): The batch of ciphertexts.
        b (PlaintextBatchArray): The batch of plaintexts.

        Returns:
        CiphertextBatchArray: The difference of the ciphertexts and plaintext.
        """
        ...

    def relinearize(
        self, ciphertexts: "CiphertextBatchArray"
    ) -> "CiphertextBatchArray":
        """
        Relinearizes the given batch of ciphertexts.

        Parameters:
        ciphertexts (CiphertextBatchArray): The batch of ciphertexts to relinearize.

        Returns:
        CiphertextBatchArray: The relinearized ciphertexts.
        """
        ...
